from datetime import datetime
from decimal import Decimal
from enum import StrEnum
from typing import Optional, Self
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Node, NodeID

from ai.backend.common.types import AutoScalingMetricSource as CommonAutoScalingMetricSource
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.data.deployment.scale import ModelDeploymentAutoScalingRuleCreator
from ai.backend.manager.data.deployment.scale_modifier import ModelDeploymentAutoScalingRuleModifier
from ai.backend.manager.data.deployment.types import ModelDeploymentAutoScalingRuleData
from ai.backend.manager.errors.common import ServerMisconfiguredError
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.create_auto_scaling_rule import (
    CreateAutoScalingRuleAction,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.delete_auto_scaling_rule import (
    DeleteAutoScalingRuleAction,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.update_auto_scaling_rule import (
    UpdateAutoScalingRuleAction,
)
from ai.backend.manager.types import OptionalState


@strawberry.enum(description="Added in 25.1.0")
class AutoScalingMetricSource(StrEnum):
    KERNEL = "KERNEL"
    INFERENCE_FRAMEWORK = "INFERENCE_FRAMEWORK"


@strawberry.type
class AutoScalingRule(Node):
    id: NodeID[str]

    metric_source: AutoScalingMetricSource = strawberry.field(
        description="Added in 25.16.0 (e.g. KERNEL, INFERENCE_FRAMEWORK)"
    )
    metric_name: str = strawberry.field()

    min_threshold: Optional[Decimal] = strawberry.field(
        description="Added in 25.16.0: The minimum threshold for scaling (e.g. 0.5)"
    )
    max_threshold: Optional[Decimal] = strawberry.field(
        description="Added in 25.16.0: The maximum threshold for scaling (e.g. 21.0)"
    )

    step_size: int = strawberry.field(
        description="Added in 25.16.0: The step size for scaling (e.g. 1)."
    )
    time_window: int = strawberry.field(
        description="Added in 25.16.0: The time window (seconds) for scaling (e.g. 60)."
    )

    min_replicas: Optional[int] = strawberry.field(
        description="Added in 25.16.0: The minimum number of replicas (e.g. 1)."
    )
    max_replicas: Optional[int] = strawberry.field(
        description="Added in 25.16.0: The maximum number of replicas (e.g. 10)."
    )

    created_at: datetime
    last_triggered_at: datetime

    @classmethod
    def from_dataclass(cls, data: ModelDeploymentAutoScalingRuleData) -> Self:
        return cls(
            id=ID(str(data.id)),
            metric_source=AutoScalingMetricSource(data.metric_source.name),
            metric_name=data.metric_name,
            min_threshold=data.min_threshold,
            max_threshold=data.max_threshold,
            step_size=data.step_size,
            time_window=data.time_window,
            min_replicas=data.min_replicas,
            max_replicas=data.max_replicas,
            created_at=data.created_at,
            last_triggered_at=data.last_triggered_at,
        )


# Input Types
@strawberry.input
class CreateAutoScalingRuleInput:
    model_deployment_id: ID
    metric_source: AutoScalingMetricSource
    metric_name: str
    min_threshold: Optional[Decimal]
    max_threshold: Optional[Decimal]
    step_size: int
    time_window: int
    min_replicas: Optional[int]
    max_replicas: Optional[int]

    def to_creator(self) -> ModelDeploymentAutoScalingRuleCreator:
        return ModelDeploymentAutoScalingRuleCreator(
            model_deployment_id=UUID(self.model_deployment_id),
            metric_source=CommonAutoScalingMetricSource(self.metric_source.lower()),
            metric_name=self.metric_name,
            min_threshold=self.min_threshold,
            max_threshold=self.max_threshold,
            step_size=self.step_size,
            time_window=self.time_window,
            min_replicas=self.min_replicas,
            max_replicas=self.max_replicas,
        )


@strawberry.input
class UpdateAutoScalingRuleInput:
    id: ID
    metric_source: Optional[AutoScalingMetricSource]
    metric_name: Optional[str]
    min_threshold: Optional[Decimal]
    max_threshold: Optional[Decimal]
    step_size: Optional[int]
    time_window: Optional[int]
    min_replicas: Optional[int]
    max_replicas: Optional[int]

    def to_action(self) -> UpdateAutoScalingRuleAction:
        optional_state_metric_source = OptionalState[CommonAutoScalingMetricSource].nop()
        if isinstance(self.metric_source, AutoScalingMetricSource):
            optional_state_metric_source = OptionalState[CommonAutoScalingMetricSource].update(
                CommonAutoScalingMetricSource(self.metric_source)
            )
        return UpdateAutoScalingRuleAction(
            auto_scaling_rule_id=UUID(self.id),
            modifier=ModelDeploymentAutoScalingRuleModifier(
                metric_source=optional_state_metric_source,
                metric_name=OptionalState[str].from_graphql(self.metric_name),
                min_threshold=OptionalState[Decimal].from_graphql(self.min_threshold),
                max_threshold=OptionalState[Decimal].from_graphql(self.max_threshold),
                step_size=OptionalState[int].from_graphql(self.step_size),
                time_window=OptionalState[int].from_graphql(self.time_window),
                min_replicas=OptionalState[int].from_graphql(self.min_replicas),
                max_replicas=OptionalState[int].from_graphql(self.max_replicas),
            ),
        )


@strawberry.input
class DeleteAutoScalingRuleInput:
    id: ID


# Payload Types
@strawberry.type
class CreateAutoScalingRulePayload:
    auto_scaling_rule: AutoScalingRule


@strawberry.type
class UpdateAutoScalingRulePayload:
    auto_scaling_rule: AutoScalingRule


@strawberry.type
class DeleteAutoScalingRulePayload:
    id: ID


@strawberry.mutation(description="Added in 25.16.0")
async def create_auto_scaling_rule(
    input: CreateAutoScalingRuleInput, info: Info[StrawberryGQLContext]
) -> CreateAutoScalingRulePayload:
    deployment_processor = info.context.processors.deployment
    if deployment_processor is None:
        raise ServerMisconfiguredError("Deployment processor is not available")
    result = await deployment_processor.create_auto_scaling_rule.wait_for_complete(
        action=CreateAutoScalingRuleAction(input.to_creator())
    )
    return CreateAutoScalingRulePayload(
        auto_scaling_rule=AutoScalingRule.from_dataclass(result.data)
    )


@strawberry.mutation(description="Added in 25.16.0")
async def update_auto_scaling_rule(
    input: UpdateAutoScalingRuleInput, info: Info[StrawberryGQLContext]
) -> UpdateAutoScalingRulePayload:
    deployment_processor = info.context.processors.deployment
    if deployment_processor is None:
        raise ServerMisconfiguredError("Deployment processor is not available")
    action_result = await deployment_processor.update_auto_scaling_rule.wait_for_complete(
        input.to_action()
    )
    return UpdateAutoScalingRulePayload(
        auto_scaling_rule=AutoScalingRule.from_dataclass(action_result.data)
    )


@strawberry.mutation(description="Added in 25.16.0")
async def delete_auto_scaling_rule(
    input: DeleteAutoScalingRuleInput, info: Info[StrawberryGQLContext]
) -> DeleteAutoScalingRulePayload:
    deployment_processor = info.context.processors.deployment
    if deployment_processor is None:
        raise ServerMisconfiguredError("Deployment processor is not available")
    _ = await deployment_processor.delete_auto_scaling_rule.wait_for_complete(
        DeleteAutoScalingRuleAction(auto_scaling_rule_id=UUID(input.id))
    )
    return DeleteAutoScalingRulePayload(id=ID(input.id))



"""Deployment service for managing model deployments."""

import logging
from datetime import datetime, timedelta
from decimal import Decimal
from uuid import uuid4

from ai.backend.common.data.model_deployment.types import (
    DeploymentStrategy,
    ModelDeploymentStatus,
)
from ai.backend.common.types import (
    AutoScalingMetricSource,
    ClusterMode,
    ResourceSlot,
    RuntimeVariant,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.deployment.types import (
    ClusterConfigData,
    DeploymentNetworkSpec,
    ExtraVFolderMountData,
    ModelDeploymentAccessTokenData,
    ModelDeploymentAutoScalingRuleData,
    ModelDeploymentData,
    ModelDeploymentMetadataInfo,
    ModelMountConfigData,
    ModelRevisionData,
    ModelRuntimeConfigData,
    ReplicaStateData,
    ResourceConfigData,
)
from ai.backend.manager.services.deployment.actions.access_token.create_access_token import (
    CreateAccessTokenAction,
    CreateAccessTokenActionResult,
)
from ai.backend.manager.services.deployment.actions.access_token.list_access_tokens import (
    ListAccessTokensAction,
    ListAccessTokensActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.batch_load_auto_scaling_rules import (
    BatchLoadAutoScalingRulesAction,
    BatchLoadAutoScalingRulesActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.create_auto_scaling_rule import (
    CreateAutoScalingRuleAction,
    CreateAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.delete_auto_scaling_rule import (
    DeleteAutoScalingRuleAction,
    DeleteAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.update_auto_scaling_rule import (
    UpdateAutoScalingRuleAction,
    UpdateAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.batch_load_deployments import (
    BatchLoadDeploymentsAction,
    BatchLoadDeploymentsActionResult,
)
from ai.backend.manager.services.deployment.actions.batch_load_replicas_by_revision_ids import (
    BatchLoadReplicasByRevisionIdsAction,
    BatchLoadReplicasByRevisionIdsActionResult,
)
from ai.backend.manager.services.deployment.actions.create_deployment import (
    CreateDeploymentAction,
    CreateDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.create_legacy_deployment import (
    CreateLegacyDeploymentAction,
    CreateLegacyDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.destroy_deployment import (
    DestroyDeploymentAction,
    DestroyDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.list_replicas import (
    ListReplicasAction,
    ListReplicasActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.add_model_revision import (
    AddModelRevisionAction,
    AddModelRevisionActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.batch_load_revisions import (
    BatchLoadRevisionsAction,
    BatchLoadRevisionsActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.create_model_revision import (
    CreateModelRevisionAction,
    CreateModelRevisionActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_deployment_id import (
    GetRevisionByDeploymentIdAction,
    GetRevisionByDeploymentIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_id import (
    GetRevisionByIdAction,
    GetRevisionByIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_replica_id import (
    GetRevisionByReplicaIdAction,
    GetRevisionByReplicaIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revisions_by_deployment_id import (
    GetRevisionsByDeploymentIdAction,
    GetRevisionsByDeploymentIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.list_revisions import (
    ListRevisionsAction,
    ListRevisionsActionResult,
)
from ai.backend.manager.services.deployment.actions.sync_replicas import (
    SyncReplicaAction,
    SyncReplicaActionResult,
)
from ai.backend.manager.services.deployment.actions.update_deployment import (
    UpdateDeploymentAction,
    UpdateDeploymentActionResult,
)
from ai.backend.manager.sokovan.deployment import DeploymentController
from ai.backend.manager.sokovan.deployment.types import DeploymentLifecycleType

log = BraceStyleAdapter(logging.getLogger(__name__))


class DeploymentService:
    """Service for managing deployments."""

    _deployment_controller: DeploymentController

    def __init__(self, deployment_controller: DeploymentController) -> None:
        """Initialize deployment service with controller."""
        self._deployment_controller = deployment_controller

    async def create_deployment(
        self, action: CreateDeploymentAction
    ) -> CreateDeploymentActionResult:
        return CreateDeploymentActionResult(
            data=ModelDeploymentData(
                id=uuid4(),
                metadata=ModelDeploymentMetadataInfo(
                    name="test-deployment",
                    status=ModelDeploymentStatus.READY,
                    tags=["tag1", "tag2"],
                    project_id=uuid4(),
                    domain_name="default",
                    created_at=datetime.now(),
                    updated_at=datetime.now(),
                ),
                network_access=DeploymentNetworkSpec(
                    open_to_public=True,
                    url="http://example.com",
                    preferred_domain_name="example.com",
                    access_token_ids=[uuid4()],
                ),
                revision_history_ids=[uuid4(), uuid4()],
                revision=mock_revision_data_1,
                scaling_rule_ids=[uuid4(), uuid4()],
                replica_state=ReplicaStateData(
                    desired_replica_count=3,
                    replica_ids=[uuid4(), uuid4(), uuid4()],
                ),
                default_deployment_strategy=DeploymentStrategy.ROLLING,
                created_user_id=uuid4(),
            )
        )

    async def create_legacy_deployment(
        self, action: CreateLegacyDeploymentAction
    ) -> CreateLegacyDeploymentActionResult:
        """Create a new legacy deployment(Model Serving).

        Args:
            action: Create legacy deployment action containing the creator specification

        Returns:
            CreateLegacyDeploymentActionResult: Result containing the created deployment info
        """
        log.info("Creating deployment with name: {}", action.draft.name)
        deployment_info = await self._deployment_controller.create_deployment(action.draft)
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_PENDING
        )
        return CreateLegacyDeploymentActionResult(data=deployment_info)

    async def update_deployment(
        self, action: UpdateDeploymentAction
    ) -> UpdateDeploymentActionResult:
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_REPLICA
        )
        return UpdateDeploymentActionResult(
            data=ModelDeploymentData(
                id=action.deployment_id,
                metadata=ModelDeploymentMetadataInfo(
                    name="test-deployment",
                    status=ModelDeploymentStatus.READY,
                    tags=["tag1", "tag2"],
                    project_id=uuid4(),
                    domain_name="default",
                    created_at=datetime.now(),
                    updated_at=datetime.now(),
                ),
                network_access=DeploymentNetworkSpec(
                    open_to_public=True,
                    url="http://example.com",
                    preferred_domain_name="example.com",
                    access_token_ids=[uuid4()],
                ),
                revision_history_ids=[uuid4(), uuid4()],
                revision=mock_revision_data_1,
                scaling_rule_ids=[uuid4(), uuid4()],
                replica_state=ReplicaStateData(
                    desired_replica_count=3,
                    replica_ids=[uuid4(), uuid4(), uuid4()],
                ),
                default_deployment_strategy=DeploymentStrategy.ROLLING,
                created_user_id=uuid4(),
            )
        )

    async def destroy_deployment(
        self, action: DestroyDeploymentAction
    ) -> DestroyDeploymentActionResult:
        """Destroy an existing deployment.

        Args:
            action: Destroy deployment action containing the endpoint ID

        Returns:
            DestroyDeploymentActionResult: Result indicating success or failure
        """
        log.info("Destroying deployment with ID: {}", action.endpoint_id)
        success = await self._deployment_controller.destroy_deployment(action.endpoint_id)
        await self._deployment_controller.mark_lifecycle_needed(DeploymentLifecycleType.DESTROYING)
        return DestroyDeploymentActionResult(success=success)

    async def batch_load_deployments(
        self, action: BatchLoadDeploymentsAction
    ) -> BatchLoadDeploymentsActionResult:
        return BatchLoadDeploymentsActionResult(
            data=[
                ModelDeploymentData(
                    id=deployment_id,
                    metadata=ModelDeploymentMetadataInfo(
                        name=f"test-deployment-{i}",
                        status=ModelDeploymentStatus.READY,
                        tags=["tag1", "tag2"],
                        project_id=uuid4(),
                        domain_name="default",
                        created_at=datetime.now(),
                        updated_at=datetime.now(),
                    ),
                    network_access=DeploymentNetworkSpec(
                        open_to_public=True,
                        url="http://example.com",
                        preferred_domain_name="example.com",
                        access_token_ids=[uuid4()],
                    ),
                    revision_history_ids=[uuid4(), uuid4()],
                    revision=mock_revision_data_1,
                    scaling_rule_ids=[uuid4(), uuid4()],
                    replica_state=ReplicaStateData(
                        desired_replica_count=3,
                        replica_ids=[uuid4(), uuid4(), uuid4()],
                    ),
                    default_deployment_strategy=DeploymentStrategy.ROLLING,
                    created_user_id=uuid4(),
                )
                for i, deployment_id in enumerate(action.deployment_ids)
            ]
        )

    async def create_auto_scaling_rule(
        self, action: CreateAutoScalingRuleAction
    ) -> CreateAutoScalingRuleActionResult:
        return CreateAutoScalingRuleActionResult(
            data=ModelDeploymentAutoScalingRuleData(
                id=uuid4(),
                model_deployment_id=action.creator.model_deployment_id,
                metric_source=action.creator.metric_source,
                metric_name=action.creator.metric_name,
                min_threshold=action.creator.min_threshold,
                max_threshold=action.creator.max_threshold,
                step_size=action.creator.step_size,
                time_window=action.creator.time_window,
                min_replicas=action.creator.min_replicas,
                max_replicas=action.creator.max_replicas,
                created_at=datetime.now(),
                last_triggered_at=datetime.now(),
            )
        )

    async def update_auto_scaling_rule(
        self, action: UpdateAutoScalingRuleAction
    ) -> UpdateAutoScalingRuleActionResult:
        return UpdateAutoScalingRuleActionResult(
            data=ModelDeploymentAutoScalingRuleData(
                id=uuid4(),
                model_deployment_id=uuid4(),
                metric_source=AutoScalingMetricSource.KERNEL,
                metric_name="test-metric",
                min_threshold=Decimal("0.5"),
                max_threshold=Decimal("21.0"),
                step_size=1,
                time_window=60,
                min_replicas=1,
                max_replicas=10,
                created_at=datetime.now(),
                last_triggered_at=datetime.now(),
            )
        )

    async def delete_auto_scaling_rule(
        self, action: DeleteAutoScalingRuleAction
    ) -> DeleteAutoScalingRuleActionResult:
        return DeleteAutoScalingRuleActionResult(success=True)

    async def create_access_token(
        self, action: CreateAccessTokenAction
    ) -> CreateAccessTokenActionResult:
        return CreateAccessTokenActionResult(
            data=ModelDeploymentAccessTokenData(
                id=uuid4(),
                token="test_token",
                valid_until=datetime.now() + timedelta(hours=1),
                created_at=datetime.now(),
            )
        )

    async def list_access_tokens(
        self, action: ListAccessTokensAction
    ) -> ListAccessTokensActionResult:
        tokens = []
        for i in range(5):
            tokens.append(
                ModelDeploymentAccessTokenData(
                    id=uuid4(),
                    token=f"test_token_{i}",
                    valid_until=datetime.now() + timedelta(hours=24 * (i + 1)),
                    created_at=datetime.now() - timedelta(hours=i),
                )
            )
        return ListAccessTokensActionResult(
            data=tokens,
            total_count=len(tokens),
        )

    async def sync_replicas(self, action: SyncReplicaAction) -> SyncReplicaActionResult:
        return SyncReplicaActionResult(success=True)

    async def add_model_revision(
        self, action: AddModelRevisionAction
    ) -> AddModelRevisionActionResult:
        return AddModelRevisionActionResult(revision=mock_revision_data_2)

    async def batch_load_auto_scaling_rules(
        self, action: BatchLoadAutoScalingRulesAction
    ) -> BatchLoadAutoScalingRulesActionResult:
        return BatchLoadAutoScalingRulesActionResult(
            data=[
                ModelDeploymentAutoScalingRuleData(
                    id=uuid4(),
                    model_deployment_id=uuid4(),
                    metric_source=AutoScalingMetricSource.KERNEL,
                    metric_name="test-metric",
                    min_threshold=Decimal("0.5"),
                    max_threshold=Decimal("21.0"),
                    step_size=1,
                    time_window=60,
                    min_replicas=1,
                    max_replicas=10,
                    created_at=datetime.now(),
                    last_triggered_at=datetime.now(),
                ),
                ModelDeploymentAutoScalingRuleData(
                    id=uuid4(),
                    model_deployment_id=uuid4(),
                    metric_source=AutoScalingMetricSource.KERNEL,
                    metric_name="test-metric",
                    min_threshold=Decimal("0.0"),
                    max_threshold=Decimal("10.0"),
                    step_size=2,
                    time_window=200,
                    min_replicas=1,
                    max_replicas=5,
                    created_at=datetime.now(),
                    last_triggered_at=datetime.now(),
                ),
            ]
        )

    async def list_replicas(self, action: ListReplicasAction) -> ListReplicasActionResult:
        return ListReplicasActionResult(
            data=[],
            total_count=0,
        )

    async def batch_load_revisions(
        self, action: BatchLoadRevisionsAction
    ) -> BatchLoadRevisionsActionResult:
        return BatchLoadRevisionsActionResult(data=[mock_revision_data_1, mock_revision_data_2])

    async def list_revisions(self, action: ListRevisionsAction) -> ListRevisionsActionResult:
        return ListRevisionsActionResult(data=[], total_count=0)

    async def get_revision_by_deployment_id(
        self, action: GetRevisionByDeploymentIdAction
    ) -> GetRevisionByDeploymentIdActionResult:
        return GetRevisionByDeploymentIdActionResult(data=mock_revision_data_1)

    async def get_revision_by_id(
        self, action: GetRevisionByIdAction
    ) -> GetRevisionByIdActionResult:
        return GetRevisionByIdActionResult(data=mock_revision_data_1)

    async def get_revision_by_replica_id(
        self, action: GetRevisionByReplicaIdAction
    ) -> GetRevisionByReplicaIdActionResult:
        return GetRevisionByReplicaIdActionResult(data=mock_revision_data_1)

    async def get_revisions_by_deployment_id(
        self, action: GetRevisionsByDeploymentIdAction
    ) -> GetRevisionsByDeploymentIdActionResult:
        # For now, return mock revision data list
        return GetRevisionsByDeploymentIdActionResult(
            data=[mock_revision_data_1, mock_revision_data_2]
        )

    async def batch_load_replicas_by_revision_ids(
        self, action: BatchLoadReplicasByRevisionIdsAction
    ) -> BatchLoadReplicasByRevisionIdsActionResult:
        # For now, return empty replica list
        return BatchLoadReplicasByRevisionIdsActionResult(data={})

    async def create_model_revision(
        self, action: CreateModelRevisionAction
    ) -> CreateModelRevisionActionResult:
        return CreateModelRevisionActionResult(revision=mock_revision_data_2)


mock_revision_data_1 = ModelRevisionData(
    id=uuid4(),
    name="test-revision",
    cluster_config=ClusterConfigData(
        mode=ClusterMode.SINGLE_NODE,
        size=1,
    ),
    resource_config=ResourceConfigData(
        resource_group_name="default",
        resource_slot=ResourceSlot.from_json({"cpu": 1, "memory": 1024}),
    ),
    model_mount_config=ModelMountConfigData(
        vfolder_id=uuid4(),
        mount_destination="/model",
        definition_path="model-definition.yaml",
    ),
    model_runtime_config=ModelRuntimeConfigData(
        runtime_variant=RuntimeVariant.VLLM,
        inference_runtime_config={"tp_size": 2, "max_length": 1024},
    ),
    extra_vfolder_mounts=[
        ExtraVFolderMountData(
            vfolder_id=uuid4(),
            mount_destination="/var",
        ),
        ExtraVFolderMountData(
            vfolder_id=uuid4(),
            mount_destination="/example",
        ),
    ],
    image_id=uuid4(),
    created_at=datetime.now(),
)

mock_revision_data_2 = ModelRevisionData(
    id=uuid4(),
    name="test-revision-2",
    cluster_config=ClusterConfigData(
        mode=ClusterMode.MULTI_NODE,
        size=1,
    ),
    resource_config=ResourceConfigData(
        resource_group_name="default",
        resource_slot=ResourceSlot.from_json({"cpu": 1, "memory": 1024}),
    ),
    model_mount_config=ModelMountConfigData(
        vfolder_id=uuid4(),
        mount_destination="/model",
        definition_path="model-definition.yaml",
    ),
    model_runtime_config=ModelRuntimeConfigData(
        runtime_variant=RuntimeVariant.NIM,
        inference_runtime_config={"tp_size": 2, "max_length": 1024},
    ),
    image_id=uuid4(),
    created_at=datetime.now(),
)



"""
Exceptions for deployment management.
"""

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class DeploymentError(BackendAIError):
    """Base exception for deployment-related errors."""

    error_type = "https://api.backend.ai/probs/deployment-failed"
    error_title = "Deployment operation failed."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.SCHEDULE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class InvalidEndpointState(DeploymentError):
    """Raised when an endpoint is in an invalid state for the requested operation."""

    error_type = "https://api.backend.ai/probs/invalid-endpoint-state"
    error_title = "Invalid endpoint state."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.BAD_REQUEST,  # Use BAD_REQUEST instead of INVALID_STATE
        )


class RouteCreationFailed(DeploymentError):
    """Raised when route creation fails."""

    error_type = "https://api.backend.ai/probs/route-creation-failed"
    error_title = "Failed to create route."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ScalingOperationFailed(DeploymentError):
    """Raised when a scaling operation fails."""

    error_type = "https://api.backend.ai/probs/scaling-failed"
    error_title = "Scaling operation failed."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ServiceInfoRetrievalFailed(DeploymentError):
    """Raised when service info cannot be retrieved."""

    error_type = "https://api.backend.ai/probs/service-info-retrieval-failed"
    error_title = "Failed to retrieve service information."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )



from __future__ import annotations

import sqlalchemy as sa

from ai.backend.manager.models.scaling_group import (
    ScalingGroupForProjectRow,
    ScalingGroupRow,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class ScalingGroupConditions:
    """Query conditions for scaling groups."""

    @staticmethod
    def by_name_contains(name: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return ScalingGroupRow.name.ilike(f"%{name}%")
            else:
                return ScalingGroupRow.name.like(f"%{name}%")

        return inner

    @staticmethod
    def by_name_equals(name: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return sa.func.lower(ScalingGroupRow.name) == name.lower()
            else:
                return ScalingGroupRow.name == name

        return inner

    @staticmethod
    def by_description_contains(description: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return ScalingGroupRow.description.ilike(f"%{description}%")
            else:
                return ScalingGroupRow.description.like(f"%{description}%")

        return inner

    @staticmethod
    def by_description_equals(description: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return sa.func.lower(ScalingGroupRow.description) == description.lower()
            else:
                return ScalingGroupRow.description == description

        return inner

    @staticmethod
    def by_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ScalingGroupRow.is_active == is_active

        return inner

    @staticmethod
    def by_is_public(is_public: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ScalingGroupRow.is_public == is_public

        return inner

    @staticmethod
    def by_driver(driver: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ScalingGroupRow.driver == driver

        return inner

    @staticmethod
    def by_scheduler(scheduler: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ScalingGroupRow.scheduler == scheduler

        return inner

    @staticmethod
    def by_use_host_network(use_host_network: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ScalingGroupRow.use_host_network == use_host_network

        return inner

    @staticmethod
    def by_project(project_id: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ScalingGroupRow.name.in_(
                sa.select(ScalingGroupForProjectRow.scaling_group).where(
                    ScalingGroupForProjectRow.group == project_id
                )
            )

        return inner

    @staticmethod
    def by_cursor_forward(cursor_name: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        ScalingGroup uses name as primary key.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(ScalingGroupRow.created_at)
                .where(ScalingGroupRow.name == cursor_name)
                .scalar_subquery()
            )
            return ScalingGroupRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_name: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        ScalingGroup uses name as primary key.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(ScalingGroupRow.created_at)
                .where(ScalingGroupRow.name == cursor_name)
                .scalar_subquery()
            )
            return ScalingGroupRow.created_at > subquery

        return inner


class ScalingGroupOrders:
    """Query orders for scaling groups."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ScalingGroupRow.name.asc()
        else:
            return ScalingGroupRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ScalingGroupRow.created_at.asc()
        else:
            return ScalingGroupRow.created_at.desc()

    @staticmethod
    def is_active(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ScalingGroupRow.is_active.asc()
        else:
            return ScalingGroupRow.is_active.desc()

    @staticmethod
    def is_public(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ScalingGroupRow.is_public.asc()
        else:
            return ScalingGroupRow.is_public.desc()



from typing import Any, Mapping, Sequence, override

from ai.backend.manager.config.loader.types import AbstractConfigLoader


class ConfigOverrider(AbstractConfigLoader):
    def __init__(
        self,
        overrides: Sequence[tuple[tuple[str, ...], Any]],
    ) -> None:
        self._overrides = overrides

    @override
    async def load(self) -> Mapping[str, Any]:
        cfg: dict[str, Any] = {}

        def set_key(path: Sequence[str], value: Any) -> None:
            node = cfg
            for key in path[:-1]:
                node = node.setdefault(key, {})
            node[path[-1]] = value

        for path, val in self._overrides:
            set_key(path, val)

        return cfg



import os
from collections.abc import Mapping, Sequence
from typing import Any, override

from ai.backend.manager.config.loader.types import AbstractConfigLoader


class EnvLoader(AbstractConfigLoader):
    _envs: list[tuple[tuple[str, ...], str]]

    def __init__(self, envs: list[tuple[tuple[str, ...], str]]) -> None:
        self._envs = envs

    @override
    async def load(self) -> Mapping[str, Any]:
        cfg: dict[str, Any] = {}

        def set_key(path: Sequence[str], value: Any) -> None:
            node = cfg
            for key in path[:-1]:
                node = node.setdefault(key, {})
            node[path[-1]] = value

        for key_path, env in self._envs:
            if (val := os.getenv(env)) is not None:
                set_key(key_path, val)

        return cfg



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.deployment.access_token import ModelDeploymentAccessTokenCreator
from ai.backend.manager.data.deployment.types import ModelDeploymentAccessTokenData
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class CreateAccessTokenAction(DeploymentBaseAction):
    creator: ModelDeploymentAccessTokenCreator

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.creator.model_deployment_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create"


@dataclass
class CreateAccessTokenActionResult(BaseActionResult):
    data: ModelDeploymentAccessTokenData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.id)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.services.resource_preset.actions.base import ResourcePresetAction
from ai.backend.manager.services.resource_preset.types import ResourcePresetCreator


@dataclass
class CreateResourcePresetAction(ResourcePresetAction):
    creator: ResourcePresetCreator

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create"


@dataclass
class CreateResourcePresetActionResult(BaseActionResult):
    resource_preset: ResourcePresetData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.resource_preset.id)


# TODO: Create exceptions.



from __future__ import annotations

import uuid
from typing import Optional, Self

import strawberry
from strawberry import ID, UNSET, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.base import to_global_id

from ...data.vfs_storage.creator import VFSStorageCreator
from ...data.vfs_storage.modifier import VFSStorageModifier
from ...data.vfs_storage.types import VFSStorageData
from ...services.vfs_storage.actions.create import CreateVFSStorageAction
from ...services.vfs_storage.actions.delete import DeleteVFSStorageAction
from ...services.vfs_storage.actions.get import GetVFSStorageAction
from ...services.vfs_storage.actions.list import ListVFSStorageAction
from ...services.vfs_storage.actions.update import UpdateVFSStorageAction
from ...types import OptionalState
from .types import StrawberryGQLContext


@strawberry.type(description="Added in 25.16.0. VFS Storage configuration")
class VFSStorage(Node):
    id: NodeID[str]
    name: str
    host: str
    base_path: str

    @classmethod
    def from_dataclass(cls, data: VFSStorageData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            host=data.host,
            base_path=str(data.base_path),
        )


VFSStorageEdge = Edge[VFSStorage]


@strawberry.type(description="Added in 25.16.0. VFS Storage connection")
class VFSStorageConnection(Connection[VFSStorage]):
    @strawberry.field
    def count(self) -> int:
        return len(self.edges)


@strawberry.field(description="Added in 25.16.0. Get a VFS storage by ID")
async def vfs_storage(id: ID, info: Info[StrawberryGQLContext]) -> Optional[VFSStorage]:
    processors = info.context.processors
    action_result = await processors.vfs_storage.get.wait_for_complete(
        GetVFSStorageAction(storage_id=uuid.UUID(id))
    )
    return VFSStorage.from_dataclass(action_result.result)


@strawberry.field(description="Added in 25.16.0. List all VFS storages")
async def vfs_storages(
    info: Info[StrawberryGQLContext],
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> VFSStorageConnection:
    processors = info.context.processors

    action_result = await processors.vfs_storage.list_storages.wait_for_complete(
        ListVFSStorageAction()
    )

    nodes = [VFSStorage.from_dataclass(data) for data in action_result.data]
    edges = [VFSStorageEdge(node=node, cursor=to_global_id(VFSStorage, node.id)) for node in nodes]

    return VFSStorageConnection(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )


@strawberry.input(description="Added in 25.16.0. Input for creating VFS storage")
class CreateVFSStorageInput:
    name: str
    host: str
    base_path: str

    def to_creator(self) -> VFSStorageCreator:
        return VFSStorageCreator(
            name=self.name,
            host=self.host,
            base_path=self.base_path,
        )


@strawberry.input(description="Added in 25.16.0. Input for updating VFS storage")
class UpdateVFSStorageInput:
    id: ID
    name: Optional[str] = UNSET
    host: Optional[str] = UNSET
    base_path: Optional[str] = UNSET

    def to_modifier(self) -> VFSStorageModifier:
        return VFSStorageModifier(
            name=OptionalState[str].from_graphql(self.name),
            host=OptionalState[str].from_graphql(self.host),
            base_path=OptionalState[str].from_graphql(self.base_path),
        )


@strawberry.input(description="Added in 25.16.0. Input for deleting VFS storage")
class DeleteVFSStorageInput:
    id: ID


@strawberry.type(description="Added in 25.16.0. Payload for creating VFS storage")
class CreateVFSStoragePayload:
    vfs_storage: VFSStorage


@strawberry.type(description="Added in 25.16.0. Payload for updating VFS storage")
class UpdateVFSStoragePayload:
    vfs_storage: VFSStorage


@strawberry.type(description="Added in 25.16.0. Payload for deleting VFS storage")
class DeleteVFSStoragePayload:
    id: ID


@strawberry.mutation(
    name="createVFSStorage", description="Added in 25.16.0. Create a new VFS storage"
)
async def create_vfs_storage(
    input: CreateVFSStorageInput, info: Info[StrawberryGQLContext]
) -> CreateVFSStoragePayload:
    processors = info.context.processors

    action_result = await processors.vfs_storage.create.wait_for_complete(
        CreateVFSStorageAction(
            creator=input.to_creator(),
        )
    )

    return CreateVFSStoragePayload(vfs_storage=VFSStorage.from_dataclass(action_result.result))


@strawberry.mutation(
    name="updateVFSStorage", description="Added in 25.16.0. Update an existing VFS storage"
)
async def update_vfs_storage(
    input: UpdateVFSStorageInput, info: Info[StrawberryGQLContext]
) -> UpdateVFSStoragePayload:
    processors = info.context.processors

    action_result = await processors.vfs_storage.update.wait_for_complete(
        UpdateVFSStorageAction(
            id=uuid.UUID(input.id),
            modifier=input.to_modifier(),
        )
    )

    return UpdateVFSStoragePayload(vfs_storage=VFSStorage.from_dataclass(action_result.result))


@strawberry.mutation(name="deleteVFSStorage", description="Added in 25.16.0. Delete a VFS storage")
async def delete_vfs_storage(
    input: DeleteVFSStorageInput, info: Info[StrawberryGQLContext]
) -> DeleteVFSStoragePayload:
    processors = info.context.processors

    action_result = await processors.vfs_storage.delete.wait_for_complete(
        DeleteVFSStorageAction(
            storage_id=uuid.UUID(input.id),
        )
    )

    return DeleteVFSStoragePayload(id=ID(str(action_result.deleted_storage_id)))



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class UpdateFullNameAction(AuthAction):
    user_id: str
    full_name: str
    domain_name: str
    email: str

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update_full_name"


@dataclass
class UpdateFullNameActionResult(BaseActionResult):
    success: bool

    @override
    def entity_id(self) -> Optional[str]:
        return None



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING, Optional, cast, override

from pydantic import Field

from ai.backend.common.bgtask.task.base import (
    BaseBackgroundTaskHandler,
    BaseBackgroundTaskManifest,
    BaseBackgroundTaskResult,
)
from ai.backend.common.data.session.types import CustomizedImageVisibilityScope
from ai.backend.common.docker import DEFAULT_KERNEL_FEATURE, KernelFeatures, LabelName
from ai.backend.common.events.event_types.bgtask.broadcast import (
    BaseBgtaskDoneEvent,
    BaseBgtaskEvent,
    BgtaskStatus,
)
from ai.backend.common.events.hub.propagators.cache import WithCachePropagator
from ai.backend.common.events.types import EventCacheDomain, EventDomain
from ai.backend.common.exception import BgtaskCancelledError, BgtaskFailedError
from ai.backend.common.types import ImageRegistry, SessionId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.bgtask.types import ManagerBgtaskName
from ai.backend.manager.data.image.types import ImageIdentifier

if TYPE_CHECKING:
    from ai.backend.common.events.fetcher import EventFetcher
    from ai.backend.common.events.hub.hub import EventHub
    from ai.backend.manager.models.image import ImageRow
    from ai.backend.manager.registry import AgentRegistry
    from ai.backend.manager.repositories.session.repository import SessionRepository

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class CommitSessionResult(BaseBackgroundTaskResult):
    """
    Result of commit session background task.
    Contains the rescanned image ID or error message.
    """

    image_id: Optional[uuid.UUID] = Field(
        default=None, description="ID of the rescanned image after commit"
    )
    error_message: Optional[str] = Field(default=None, description="Error message if task failed")


class CommitSessionManifest(BaseBackgroundTaskManifest):
    """
    Manifest for committing a session and uploading it as a customized image.
    """

    session_id: SessionId = Field(description="Session ID to commit")
    registry_hostname: str = Field(description="Registry hostname to push the image")
    registry_project: str = Field(description="Registry project name")
    image_name: str = Field(description="Name for the customized image")
    image_visibility: CustomizedImageVisibilityScope = Field(
        description="Visibility scope of the customized image"
    )
    image_owner_id: str = Field(description="Owner ID of the customized image")
    user_email: str = Field(description="User email for customized image metadata")


class CommitSessionHandler(BaseBackgroundTaskHandler[CommitSessionManifest, CommitSessionResult]):
    """
    Background task handler for committing sessions and uploading as customized images.
    """

    _session_repository: SessionRepository
    _agent_registry: AgentRegistry
    _event_hub: EventHub
    _event_fetcher: EventFetcher

    def __init__(
        self,
        session_repository: SessionRepository,
        agent_registry: AgentRegistry,
        event_hub: EventHub,
        event_fetcher: EventFetcher,
    ) -> None:
        self._session_repository = session_repository
        self._agent_registry = agent_registry
        self._event_hub = event_hub
        self._event_fetcher = event_fetcher

    @classmethod
    @override
    def name(cls) -> ManagerBgtaskName:
        return ManagerBgtaskName.COMMIT_SESSION  # type: ignore[return-value]

    @classmethod
    @override
    def manifest_type(cls) -> type[CommitSessionManifest]:
        return CommitSessionManifest

    @override
    async def execute(self, manifest: CommitSessionManifest) -> CommitSessionResult:
        try:
            # Get session and validate
            session = await self._session_repository.get_session_by_id(manifest.session_id)
            if not session:
                error_msg = f"Session {manifest.session_id} not found"
                log.error(error_msg)
                return CommitSessionResult(error_message=error_msg)

            # Get registry configuration
            registry_conf = await self._session_repository.get_container_registry(
                manifest.registry_hostname, manifest.registry_project
            )
            if not registry_conf:
                error_msg = f"Project {manifest.registry_project} not found in registry {manifest.registry_hostname}"
                log.error(error_msg)
                return CommitSessionResult(error_message=error_msg)

            # Resolve base image
            image_row = await self._session_repository.resolve_image([
                ImageIdentifier(session.main_kernel.image, session.main_kernel.architecture)
            ])
            base_image_ref = image_row.image_ref

            # Build new image canonical name
            filtered_tag_set = [
                x for x in base_image_ref.tag.split("-") if not x.startswith("customized_")
            ]

            if base_image_ref.name == "":
                new_name = base_image_ref.project
            else:
                new_name = base_image_ref.name

            new_canonical = f"{manifest.registry_hostname}/{manifest.registry_project}/{new_name}:{'-'.join(filtered_tag_set)}"

            # Check for existing customized image
            existing_row = await self._session_repository.get_existing_customized_image(
                new_canonical,
                manifest.image_visibility.value,
                manifest.image_owner_id,
                manifest.image_name,
            )

            customized_image_id: str
            kern_features: list[str]
            if existing_row is not None:
                existing_image: ImageRow = existing_row
                labels = existing_image.labels or {}
                kern_features_str = labels.get(LabelName.FEATURES, DEFAULT_KERNEL_FEATURE)
                kern_features = (
                    kern_features_str.split() if kern_features_str else [DEFAULT_KERNEL_FEATURE]
                )
                customized_image_id = labels.get(LabelName.CUSTOMIZED_ID, str(uuid.uuid4()))
                log.debug("reusing existing customized image ID {}", customized_image_id)
            else:
                kern_features = [DEFAULT_KERNEL_FEATURE]
                customized_image_id = str(uuid.uuid4())
                # Remove PRIVATE label for customized images
                kern_features = [
                    feat for feat in kern_features if feat != KernelFeatures.PRIVATE.value
                ]

            new_canonical += f"-customized_{customized_image_id.replace('-', '')}"

            from ai.backend.common.docker import ImageRef

            new_image_ref = ImageRef.from_image_str(
                new_canonical,
                None,
                manifest.registry_hostname,
                architecture=base_image_ref.architecture,
                is_local=base_image_ref.is_local,
            )

            # Prepare image labels
            image_labels: dict[str | LabelName, str] = {
                LabelName.CUSTOMIZED_OWNER: f"{manifest.image_visibility.value}:{manifest.image_owner_id}",
                LabelName.CUSTOMIZED_NAME: manifest.image_name,
                LabelName.CUSTOMIZED_ID: customized_image_id,
                LabelName.FEATURES: " ".join(kern_features),
            }
            match manifest.image_visibility:
                case CustomizedImageVisibilityScope.USER:
                    image_labels[LabelName.CUSTOMIZED_USER_EMAIL] = manifest.user_email

            # Commit session
            log.info("Committing session {}", manifest.session_id)
            resp = await self._agent_registry.commit_session(
                session,
                new_image_ref,
                extra_labels=image_labels,
            )
            bgtask_id = cast(uuid.UUID, resp["bgtask_id"])

            # Wait for commit to complete
            await self._wait_for_agent_bgtask(bgtask_id, "Commit")

            # Push image to registry if not local
            if not new_image_ref.is_local:
                log.info("Pushing image to registry")
                image_registry = ImageRegistry(
                    name=manifest.registry_hostname,
                    url=str(registry_conf.url),
                    username=registry_conf.username,
                    password=registry_conf.password,
                )
                resp = await self._agent_registry.push_image(
                    session.main_kernel.agent,
                    new_image_ref,
                    image_registry,
                )
                bgtask_id = cast(uuid.UUID, resp["bgtask_id"])
                await self._wait_for_agent_bgtask(bgtask_id, "Push")

            # Rescan updated image
            log.info("Rescanning image")
            rescan_result = await self._session_repository.rescan_images(
                new_image_ref.canonical,
                manifest.registry_project,
                reporter=None,
            )

            if len(rescan_result.images) == 0:
                rescan_errors = ",".join(rescan_result.errors)
                error_msg = (
                    f"Session commit succeeded, but no image was rescanned, Error: {rescan_errors}"
                )
                log.error(error_msg)
                return CommitSessionResult(error_message=error_msg)
            elif len(rescan_result.images) > 1:
                log.warning(
                    "More than two images were rescanned unexpectedly. Rescanned Images: {}",
                    rescan_result.images,
                )

            result_image_id = rescan_result.images[0].id
            log.info("Session commit completed successfully. Image ID: {}", result_image_id)
            return CommitSessionResult(image_id=result_image_id)

        except Exception as e:
            log.exception("Failed to commit session {}", manifest.session_id)
            return CommitSessionResult(error_message=str(e))

    async def _wait_for_agent_bgtask(self, bgtask_id: uuid.UUID, operation_name: str) -> None:
        """Wait for an agent background task to complete."""
        propagator = WithCachePropagator(self._event_fetcher)
        self._event_hub.register_event_propagator(
            propagator, [(EventDomain.BGTASK, str(bgtask_id))]
        )
        try:
            cache_id = EventCacheDomain.BGTASK.cache_id(str(bgtask_id))
            async for event in propagator.receive(cache_id):
                if not isinstance(event, BaseBgtaskEvent):
                    log.warning("unexpected event: {}", event)
                    continue
                match event.status():
                    case BgtaskStatus.DONE | BgtaskStatus.PARTIAL_SUCCESS:
                        log.info("{} completed", operation_name)
                        return
                    case BgtaskStatus.FAILED:
                        error_msg = cast(BaseBgtaskDoneEvent, event).message
                        log.error("{} failed: {}", operation_name, error_msg)
                        raise BgtaskFailedError(extra_msg=error_msg)
                    case BgtaskStatus.CANCELLED:
                        log.warning("{} cancelled", operation_name)
                        raise BgtaskCancelledError(extra_msg="Operation cancelled")
                    case BgtaskStatus.UPDATED:
                        continue
                    case _:
                        log.warning("unexpected bgtask done event: {}", event)
        finally:
            self._event_hub.unregister_event_propagator(propagator.id())



"""Action for creating legacy deployments(Model Service)."""

from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.deployment.creator import DeploymentCreationDraft
from ai.backend.manager.data.deployment.types import DeploymentInfo
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class CreateLegacyDeploymentAction(DeploymentBaseAction):
    """Action to create a new legacy deployment(Model Service)."""

    draft: DeploymentCreationDraft

    @override
    def entity_id(self) -> Optional[str]:
        return None  # New deployment doesn't have an ID yet

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create"


@dataclass
class CreateLegacyDeploymentActionResult(BaseActionResult):
    data: DeploymentInfo

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.id)



from __future__ import annotations

import logging
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Optional, Self
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import relationship

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.services.resource_preset.types import (
    ResourcePresetCreator,
)

from .base import (
    Base,
    IDColumn,
    ResourceSlotColumn,
)

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.models"))

__all__: Sequence[str] = ("resource_presets",)


type QueryStatement = sa.sql.Select


def filter_by_name(name: str) -> Callable[[QueryStatement], QueryStatement]:
    return lambda query_stmt: query_stmt.where(ResourcePresetRow.name == name)


def filter_by_id(id: UUID) -> Callable[[QueryStatement], QueryStatement]:
    return lambda query_stmt: query_stmt.where(ResourcePresetRow.id == id)


QueryOption = Callable[[Any], Callable[[QueryStatement], QueryStatement]]


class ResourcePresetRow(Base):
    __tablename__ = "resource_presets"
    id = IDColumn()
    name = sa.Column("name", sa.String(length=256), nullable=False)
    resource_slots = sa.Column("resource_slots", ResourceSlotColumn(), nullable=False)
    shared_memory = sa.Column("shared_memory", sa.BigInteger(), nullable=True)

    # If `scaling_group_name` is None, the preset is global
    scaling_group_name = sa.Column(
        "scaling_group_name", sa.String(length=64), nullable=True, server_default=sa.null()
    )
    scaling_group_row = relationship(
        "ScalingGroupRow",
        back_populates="resource_preset_rows",
        primaryjoin="ScalingGroupRow.name == foreign(ResourcePresetRow.scaling_group_name)",
    )

    __table_args__ = (
        sa.Index(
            "ix_resource_presets_name_null_scaling_group_name",
            name,
            postgresql_where=scaling_group_name.is_(None),
            unique=True,
        ),
        sa.Index(
            "ix_resource_presets_name_scaling_group_name",
            name,
            scaling_group_name,
            postgresql_where=scaling_group_name.isnot(None),
            unique=True,
        ),
    )

    @classmethod
    async def create(
        cls,
        creator: ResourcePresetCreator,
        *,
        db_session: AsyncSession,
    ) -> Optional[Self]:
        to_store = creator.fields_to_store()
        insert_stmt = sa.insert(ResourcePresetRow).values(to_store).returning(ResourcePresetRow)
        stmt = sa.select(ResourcePresetRow).from_statement(insert_stmt)

        try:
            return await db_session.scalar(stmt)
        except sa.exc.IntegrityError:
            # A resource preset with the given name and scaling group name already exists
            return None

    @classmethod
    async def update(
        cls,
        query_option: QueryOption,
        data: Mapping[str, Any],
        *,
        db_session: AsyncSession,
    ) -> Optional[Self]:
        update_stmt = sa.update(ResourcePresetRow).values(data).returning(ResourcePresetRow)
        update_stmt = query_option(update_stmt)
        stmt = (
            sa.select(ResourcePresetRow)
            .from_statement(update_stmt)
            .execution_options(populate_existing=True)
        )
        try:
            return await db_session.scalar(stmt)
        except sa.exc.IntegrityError:
            return None

    @classmethod
    async def delete(
        cls,
        query_option: QueryOption,
        *,
        db_session: AsyncSession,
    ) -> None:
        delete_stmt = sa.delete(ResourcePresetRow)
        delete_stmt = query_option(delete_stmt)
        return await db_session.execute(delete_stmt)

    def to_dataclass(self) -> ResourcePresetData:
        return ResourcePresetData(
            id=self.id,
            name=self.name,
            resource_slots=self.resource_slots,
            shared_memory=self.shared_memory,
            scaling_group_name=self.scaling_group_name,
        )


# For compatibility
resource_presets = ResourcePresetRow.__table__



"""Add `status` Column to the `Image` table, and `ImageRow` unique constraint

Revision ID: c002140f14d3
Revises: 8f85e9d0bd4e
Create Date: 2025-02-10 03:22:31.611405

"""

import enum

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import StrEnumType

# revision identifiers, used by Alembic.
revision = "c002140f14d3"
down_revision = "8f85e9d0bd4e"
branch_labels = None
depends_on = None


class ImageStatus(enum.StrEnum):
    ALIVE = "ALIVE"
    DELETED = "DELETED"


def upgrade() -> None:
    op.add_column(
        "images",
        sa.Column(
            "status", StrEnumType(ImageStatus), server_default=ImageStatus.ALIVE, nullable=False
        ),
    )
    op.create_unique_constraint(
        "uq_image_identifier", "images", ["registry", "project", "name", "tag", "architecture"]
    )


def downgrade() -> None:
    op.drop_column("images", "status")
    op.drop_constraint("uq_image_identifier", "images", type_="unique")



"""
Authentication and user-related exceptions.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)

from .common import ObjectNotFound


class InvalidCredentials(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-credentials"
    error_title = "Invalid credentials for authentication."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InsufficientPrivilege(BackendAIError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/insufficient-privilege"
    error_title = "Insufficient privilege."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class InvalidAuthParameters(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-auth-params"
    error_title = "Missing or invalid authorization parameters."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class AuthorizationFailed(BackendAIError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/auth-failed"
    error_title = "Credential/signature mismatch."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.UNAUTHORIZED,
        )


class PasswordExpired(BackendAIError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/password-expired"
    error_title = "Password has expired."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.DATA_EXPIRED,
        )


class EmailAlreadyExistsError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/email-already-exists"
    error_title = "Email already exists."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class UserCreationError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/user-creation-failed"
    error_title = "Failed to create user account."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UserNotFound(ObjectNotFound):
    object_name = "user"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class GroupMembershipNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/group-membership-not-found"
    error_title = "User is not a member of the specified group."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.GROUP,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class InvalidClientIPConfig(BackendAIError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/invalid-client-ip-config"
    error_title = "Invalid client IP configuration."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.FORBIDDEN,
        )



"""
Hook for inference session type.
Handles model serving operations like route creation and deletion.
"""

import logging
import uuid
from typing import Optional

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.model_serving.anycast import (
    EndpointRouteListUpdatedEvent,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.repositories.deployment.repository import DeploymentRepository

from ..types import SessionTransitionData
from .base import AbstractSessionHook

log = BraceStyleAdapter(logging.getLogger(__name__))


class InferenceSessionHook(AbstractSessionHook):
    _repository: DeploymentRepository
    _event_producer: EventProducer

    def __init__(
        self,
        repository: DeploymentRepository,
        event_producer: EventProducer,
    ) -> None:
        self._repository = repository
        self._event_producer = event_producer

    async def _get_endpoint_id_from_session(
        self, session: SessionTransitionData
    ) -> Optional[uuid.UUID]:
        """
        Extract endpoint ID from session by looking up the associated route.

        Args:
            session: Session transition data

        Returns:
            Endpoint ID if found, None otherwise
        """
        try:
            endpoint_id = await self._repository.get_endpoint_id_by_session(session.session_id)
            if not endpoint_id:
                log.warning(
                    "No endpoint ID found for session {}",
                    session.session_id,
                )
            return endpoint_id
        except Exception as e:
            log.error(
                "Error getting endpoint ID for session {}: {}",
                session.session_id,
                e,
            )
            return None

    async def on_transition_to_running(self, session: SessionTransitionData) -> None:
        """Handle route creation when inference session starts running."""
        log.info(
            "Creating model service route for inference session {}",
            session.session_id,
        )

        # Get endpoint ID from session metadata
        endpoint_id = await self._get_endpoint_id_from_session(session)
        if not endpoint_id:
            log.warning(
                "No endpoint ID found for inference session {}, skipping route update",
                session.session_id,
            )
            return

        try:
            # Update route info in Redis
            await self._repository.update_endpoint_route_info(endpoint_id)

            # Send event to app proxy
            await self._event_producer.anycast_event(EndpointRouteListUpdatedEvent(endpoint_id))

            log.info(
                "Successfully updated route info and notified app proxy for endpoint {} (session {})",
                endpoint_id,
                session.session_id,
            )
        except Exception as e:
            log.exception(
                "Unexpected error updating route info for endpoint {} (session {}): {}",
                endpoint_id,
                session.session_id,
                e,
            )
            # Don't fail the session transition, just log the error

    async def on_transition_to_terminated(self, session: SessionTransitionData) -> None:
        """Handle route deletion when inference session terminates."""
        log.info(
            "Deleting model service route for inference session {}",
            session.session_id,
        )

        # Get endpoint ID from session metadata
        endpoint_id = await self._get_endpoint_id_from_session(session)
        if not endpoint_id:
            log.warning(
                "No endpoint ID found for inference session {}, skipping route update",
                session.session_id,
            )
            return

        try:
            # Update route info in Redis (removal)
            await self._repository.update_endpoint_route_info(endpoint_id)

            # Send event to app proxy
            await self._event_producer.anycast_event(EndpointRouteListUpdatedEvent(endpoint_id))

            log.info(
                "Successfully updated route info and notified app proxy of route removal for endpoint {} (session {})",
                endpoint_id,
                session.session_id,
            )
        except Exception as e:
            log.exception(
                "Unexpected error updating route info for endpoint {} (session {}): {}",
                endpoint_id,
                session.session_id,
                e,
            )
            # Don't fail the session transition, just log the error



import decimal
import logging
from typing import Any

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.model_serving.types import RequesterCtx
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.repositories.model_serving.admin_repository import (
    AdminModelServingRepository,
)
from ai.backend.manager.repositories.model_serving.repository import ModelServingRepository
from ai.backend.manager.services.model_serving.actions.create_auto_scaling_rule import (
    CreateEndpointAutoScalingRuleAction,
    CreateEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_auto_scaling_rule import (
    DeleteEndpointAutoScalingRuleAction,
    DeleteEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.modify_auto_scaling_rule import (
    ModifyEndpointAutoScalingRuleAction,
    ModifyEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.scale_service_replicas import (
    ScaleServiceReplicasAction,
    ScaleServiceReplicasActionResult,
)
from ai.backend.manager.services.model_serving.exceptions import (
    EndpointAutoScalingRuleNotFound,
    EndpointNotFound,
    GenericForbidden,
    InvalidAPIParameters,
    ModelServiceNotFound,
)

log = BraceStyleAdapter(logging.getLogger(__name__))


class AutoScalingService:
    _repository: ModelServingRepository
    _admin_repository: AdminModelServingRepository

    def __init__(
        self,
        repository: ModelServingRepository,
        admin_repository: AdminModelServingRepository,
    ) -> None:
        self._repository = repository
        self._admin_repository = admin_repository

    async def check_requester_access(self, requester_ctx: RequesterCtx) -> None:
        if requester_ctx.is_authorized is False:
            raise GenericForbidden("Only authorized requests may have access key scopes.")

    async def scale_service_replicas(
        self, action: ScaleServiceReplicasAction
    ) -> ScaleServiceReplicasActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
            if not endpoint_data:
                raise ModelServiceNotFound

            success = await self._admin_repository.update_endpoint_replicas_force(
                action.service_id, action.to
            )
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )
            if not endpoint_data:
                raise ModelServiceNotFound

            success = await self._repository.update_endpoint_replicas_validated(
                action.service_id,
                action.to,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not success:
            raise ModelServiceNotFound

        return ScaleServiceReplicasActionResult(
            current_route_count=len(endpoint_data.routings) if endpoint_data.routings else 0,
            target_count=action.to,
        )

    async def create_endpoint_auto_scaling_rule(
        self, action: CreateEndpointAutoScalingRuleAction
    ) -> CreateEndpointAutoScalingRuleActionResult:
        try:
            _threshold = decimal.Decimal(action.creator.threshold)
        except decimal.InvalidOperation:
            raise InvalidAPIParameters(f"Cannot convert {action.creator.threshold} to Decimal")

        # Create auto scaling rule with access validation
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            created_rule = await self._admin_repository.create_auto_scaling_rule_force(
                endpoint_id=action.endpoint_id,
                metric_source=action.creator.metric_source,
                metric_name=action.creator.metric_name,
                threshold=_threshold,
                comparator=action.creator.comparator,
                step_size=action.creator.step_size,
                cooldown_seconds=action.creator.cooldown_seconds,
                min_replicas=action.creator.min_replicas,
                max_replicas=action.creator.max_replicas,
            )
        else:
            created_rule = await self._repository.create_auto_scaling_rule_validated(
                user_id=action.requester_ctx.user_id,
                user_role=action.requester_ctx.user_role,
                domain_name=action.requester_ctx.domain_name,
                endpoint_id=action.endpoint_id,
                metric_source=action.creator.metric_source,
                metric_name=action.creator.metric_name,
                threshold=_threshold,
                comparator=action.creator.comparator,
                step_size=action.creator.step_size,
                cooldown_seconds=action.creator.cooldown_seconds,
                min_replicas=action.creator.min_replicas,
                max_replicas=action.creator.max_replicas,
            )

        if created_rule is None:
            raise EndpointNotFound

        return CreateEndpointAutoScalingRuleActionResult(
            success=True,
            data=created_rule,
        )

    async def modify_endpoint_auto_scaling_rule(
        self, action: ModifyEndpointAutoScalingRuleAction
    ) -> ModifyEndpointAutoScalingRuleActionResult:
        fields_to_update: dict[str, Any] = action.modifier.fields_to_update()

        # Update auto scaling rule with access validation
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            updated_rule = await self._admin_repository.update_auto_scaling_rule_force(
                action.id, fields_to_update
            )
        else:
            updated_rule = await self._repository.update_auto_scaling_rule_validated(
                action.id,
                fields_to_update,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if updated_rule is None:
            raise EndpointAutoScalingRuleNotFound

        return ModifyEndpointAutoScalingRuleActionResult(
            success=True,
            data=updated_rule,
        )

    async def delete_endpoint_auto_scaling_rule(
        self, action: DeleteEndpointAutoScalingRuleAction
    ) -> DeleteEndpointAutoScalingRuleActionResult:
        # Delete auto scaling rule with access validation
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            success = await self._admin_repository.delete_auto_scaling_rule_force(action.id)
        else:
            success = await self._repository.delete_auto_scaling_rule_validated(
                action.id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not success:
            raise EndpointAutoScalingRuleNotFound

        return DeleteEndpointAutoScalingRuleActionResult(
            success=True,
        )



from __future__ import annotations

import asyncio
import functools
import json
import logging
from contextlib import AbstractAsyncContextManager as AbstractAsyncCtxMgr
from contextlib import asynccontextmanager as actxmgr
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Concatenate,
    Mapping,
    Optional,
    ParamSpec,
    Tuple,
    TypeAlias,
    TypeVar,
    cast,
    overload,
)

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as psql
from sqlalchemy.engine import create_engine as _create_engine
from sqlalchemy.exc import DBAPIError
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncEngine as SAEngine
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import sessionmaker
from tenacity import (
    AsyncRetrying,
    AttemptManager,
    RetryError,
    TryAgain,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from yarl import URL

from ai.backend.common.exception import DatabaseError
from ai.backend.common.json import ExtendedJSONEncoder
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.bootstrap import DatabaseConfig
from ai.backend.manager.errors.resource import DBOperationFailed

from ..defs import LockID
from ..types import Sentinel

if TYPE_CHECKING:
    from ai.backend.manager.config.bootstrap import BootstrapConfig


log = BraceStyleAdapter(logging.getLogger(__spec__.name))
column_constraints = ["nullable", "index", "unique", "primary_key"]

# TODO: Implement begin(), begin_readonly() for AsyncSession also


class ExtendedAsyncSAEngine(SAEngine):
    """
    A subclass to add a few more convenience methods to the SQLAlchemy's async engine.
    """

    def __init__(self, *args, **kwargs) -> None:
        self._txn_concurrency_threshold = kwargs.pop("_txn_concurrency_threshold", 0)
        self.lock_conn_timeout: float | None = (
            kwargs.pop("_lock_conn_timeout", 0) or None
        )  # Convert 0 to `None`
        super().__init__(*args, **kwargs)
        self._readonly_txn_count = 0
        self._generic_txn_count = 0
        self._sess_factory = sessionmaker(self, expire_on_commit=False, class_=SASession)
        self._readonly_sess_factory = sessionmaker(self, class_=SASession)

    def _check_generic_txn_cnt(self) -> None:
        if (
            self._txn_concurrency_threshold > 0
            and self._generic_txn_count >= self._txn_concurrency_threshold
        ):
            log.warning(
                "The number of concurrent generic transactions ({}) "
                "looks too high (warning threshold: {}).",
                self._generic_txn_count,
                self._txn_concurrency_threshold,
                stack_info=False,
            )

    def _check_readonly_txn_cnt(self) -> None:
        if (
            self._txn_concurrency_threshold > 0
            and self._readonly_txn_count >= self._txn_concurrency_threshold
        ):
            log.warning(
                "The number of concurrent read-only transactions ({}) "
                "looks too high (warning threshold: {}).",
                self._readonly_txn_count,
                self._txn_concurrency_threshold,
                stack_info=False,
            )

    @actxmgr
    async def _begin(self, connection: SAConnection) -> AsyncIterator[SAConnection]:
        """
        Begin generic transaction within the given connection.
        """
        conn_with_exec_opts = await connection.execution_options(
            postgresql_readonly=False,
        )
        async with conn_with_exec_opts.begin():
            self._generic_txn_count += 1
            self._check_generic_txn_cnt()
            try:
                yield connection
            finally:
                self._generic_txn_count -= 1

    @actxmgr
    async def _begin_readonly(
        self, connection: SAConnection, deferrable: bool = False
    ) -> AsyncIterator[SAConnection]:
        """
        Begin read-only transaction within the given connection.
        """
        conn_with_exec_opts = await connection.execution_options(
            postgresql_readonly=True,
            postgresql_deferrable=deferrable,
        )
        async with conn_with_exec_opts.begin():
            self._readonly_txn_count += 1
            self._check_readonly_txn_cnt()
            try:
                yield conn_with_exec_opts
            finally:
                self._readonly_txn_count -= 1

    @actxmgr
    async def begin(self, bind: SAConnection | None = None) -> AsyncIterator[SAConnection]:
        if bind is None:
            async with self.connect() as _bind:
                async with self._begin(_bind) as conn:
                    yield conn
        else:
            async with self._begin(bind) as conn:
                yield conn

    @actxmgr
    async def begin_read_committed(self) -> AsyncIterator[SAConnection]:
        """
        Begin a read-write connection with READ COMMITTED isolation level.
        """
        async with self.connect() as conn:
            # Set isolation level to READ COMMITTED
            conn = cast(SAConnection, conn)
            conn_with_isolation = await conn.execution_options(isolation_level="READ COMMITTED")
            async with conn_with_isolation.begin():
                yield conn_with_isolation

    async def ping(self) -> None:
        """
        Ping the database to check if the connection is alive.

        Raises:
            DatabaseError: If the ping fails or connection is not available
        """
        async with self.begin_readonly_read_committed() as conn:
            result = await conn.execute(sa.text("SELECT 1"))
            scalar_result = result.scalar()
            if scalar_result != 1:
                raise DatabaseError("Database ping failed: unexpected result")

    @actxmgr
    async def begin_readonly(
        self,
        bind: SAConnection | None = None,
        *,
        deferrable: bool = False,
        isolation_level: str | None = None,  # override only when existing conn is not given
    ) -> AsyncIterator[SAConnection]:
        if bind is None:
            async with self.connect() as _bind:
                if isolation_level is not None:
                    _bind = await _bind.execution_options(
                        isolation_level=isolation_level,
                    )
                async with self._begin_readonly(_bind, deferrable) as conn:
                    yield conn
        else:
            async with self._begin_readonly(bind, deferrable) as conn:
                yield conn

    @actxmgr
    async def begin_readonly_read_committed(self) -> AsyncIterator[SAConnection]:
        """
        Begin a read-only connection with READ COMMITTED isolation level.
        """
        async with self.connect() as conn:
            # Set isolation level to READ COMMITTED
            conn = cast(SAConnection, conn)
            conn_with_isolation = await conn.execution_options(
                isolation_level="READ COMMITTED",
                postgresql_readonly=True,
            )
            async with conn_with_isolation.begin():
                yield conn_with_isolation

    @actxmgr
    async def begin_session(
        self,
        bind: SAConnection | None = None,
        expire_on_commit: bool = False,
        commit_on_end: bool = True,
    ) -> AsyncIterator[SASession]:
        @actxmgr
        async def _begin_session(connection: SAConnection) -> AsyncIterator[SASession]:
            async with self._begin(connection) as conn:
                self._sess_factory.configure(bind=conn, expire_on_commit=expire_on_commit)
                session = self._sess_factory()
                yield session
                if commit_on_end:
                    await session.commit()

        if bind is None:
            async with self.connect() as _bind:
                async with _begin_session(_bind) as sess:
                    yield sess
        else:
            async with _begin_session(bind) as sess:
                yield sess

    @actxmgr
    async def begin_session_read_committed(self) -> AsyncIterator[SASession]:
        """
        Begin a read-write session with READ COMMITTED isolation level.
        """
        async with self.connect() as conn:
            # Set isolation level to READ COMMITTED
            conn = cast(SAConnection, conn)
            conn_with_isolation = await conn.execution_options(isolation_level="READ COMMITTED")
            async with conn_with_isolation.begin():
                # Configure session factory with the connection
                self._sess_factory.configure(bind=conn_with_isolation, expire_on_commit=False)
                session = self._sess_factory()
                session = cast(SASession, session)
                yield session
                await session.commit()

    @actxmgr
    async def begin_readonly_session(
        self,
        bind: SAConnection | None = None,
        *,
        deferrable: bool = False,
        isolation_level: str | None = None,  # override only when existing conn is not given
    ) -> AsyncIterator[SASession]:
        @actxmgr
        async def _begin_session(connection: SAConnection) -> AsyncIterator[SASession]:
            async with self._begin_readonly(connection, deferrable) as conn:
                self._readonly_sess_factory.configure(bind=conn)
                session = self._readonly_sess_factory()
                yield session

        if bind is None:
            async with self.connect() as _bind:
                if isolation_level is not None:
                    _bind = await _bind.execution_options(
                        isolation_level=isolation_level,
                    )
                async with _begin_session(_bind) as sess:
                    yield sess
        else:
            async with _begin_session(bind) as sess:
                yield sess

    @actxmgr
    async def begin_readonly_session_read_committed(self) -> AsyncIterator[SASession]:
        """
        Begin a read-only session with READ COMMITTED isolation level.
        """
        async with self.connect() as conn:
            # Set isolation level to READ COMMITTED and readonly mode
            conn = cast(SAConnection, conn)
            conn_with_isolation = await conn.execution_options(
                isolation_level="READ COMMITTED",
                postgresql_readonly=True,
            )
            async with conn_with_isolation.begin():
                # Configure session factory with the connection
                self._readonly_sess_factory.configure(bind=conn_with_isolation)
                session = self._readonly_sess_factory()
                session = cast(SASession, session)
                yield session

    @actxmgr
    async def advisory_lock(self, lock_id: LockID) -> AsyncIterator[None]:
        lock_acquired = False
        # Here we use the session-level advisory lock,
        # which follows the lifetime of underlying DB connection.
        # As such, we should keep using one single connection for both lock and unlock ops.
        async with self.connect() as lock_conn:
            try:
                # It is usually a BAD practice to directly interpolate strings into SQL statements,
                # but in this case:
                #  - The lock ID is only given from trusted codes.
                #  - asyncpg does not support parameter interpolation with raw SQL statements.
                async with asyncio.timeout(self.lock_conn_timeout):
                    await lock_conn.exec_driver_sql(
                        f"SELECT pg_advisory_lock({lock_id:d});",
                    )
                    lock_acquired = True
                    yield
            except sa.exc.DBAPIError as e:
                if getattr(e.orig, "pgcode", None) == "55P03":  # lock not available error
                    # This may happen upon shutdown after some time.
                    raise asyncio.CancelledError()
                raise
            except asyncio.CancelledError:
                raise
            finally:
                if lock_acquired and not lock_conn.closed:
                    try:
                        await lock_conn.exec_driver_sql(
                            f"SELECT pg_advisory_unlock({lock_id:d})",
                        )
                    except sa.exc.InterfaceError:
                        log.warning(
                            f"DB Connnection for lock(id: {lock_id:d}) has already been closed. Skip unlock"
                        )


P = ParamSpec("P")
TQueryResult = TypeVar("TQueryResult")


@overload
async def execute_with_txn_retry(
    txn_func: Callable[Concatenate[SASession, P], Awaitable[TQueryResult]],
    begin_trx: Callable[..., AbstractAsyncCtxMgr[SASession]],
    connection: SAConnection,
    *args: P.args,
    **kwargs: P.kwargs,
) -> TQueryResult: ...


# Setting "type ignore" here becuase Mypy deduces all fields and attributes in `sqlalchemy` module to `Any` type
# including `SASession` and `SAConnection`.
@overload
async def execute_with_txn_retry(  # type: ignore[misc]
    txn_func: Callable[Concatenate[SAConnection, P], Awaitable[TQueryResult]],
    begin_trx: Callable[..., AbstractAsyncCtxMgr[SAConnection]],
    connection: SAConnection,
    *args: P.args,
    **kwargs: P.kwargs,
) -> TQueryResult: ...


# TODO: Allow `SASession` parameter only, remove type overloading and remove `begin_trx` after migrating Core APIs to ORM APIs.
async def execute_with_txn_retry(
    txn_func: Callable[Concatenate[SASession, P], Awaitable[TQueryResult]]
    | Callable[Concatenate[SAConnection, P], Awaitable[TQueryResult]],
    begin_trx: Callable[..., AbstractAsyncCtxMgr[SASession]]
    | Callable[..., AbstractAsyncCtxMgr[SAConnection]],
    connection: SAConnection,
    *args: P.args,
    **kwargs: P.kwargs,
) -> TQueryResult:
    """
    Execute DB related function by retrying transaction in a given connection.

    The transaction retry resolves Postgres's Serialization error.
    Reference: https://www.postgresql.org/docs/current/mvcc-serialization-failure-handling.html
    """

    result: TQueryResult | Sentinel = Sentinel.TOKEN
    max_attempts = 10
    try:
        async for attempt in AsyncRetrying(
            wait=wait_exponential(multiplier=0.02, min=0.02, max=1.0),
            stop=stop_after_attempt(max_attempts),
            retry=retry_if_exception_type(TryAgain),
        ):
            with attempt:
                try:
                    async with begin_trx(bind=connection) as session_or_conn:
                        result = await txn_func(session_or_conn, *args, **kwargs)
                except DBAPIError as e:
                    if is_db_retry_error(e):
                        raise TryAgain
                    raise
    except RetryError:
        raise asyncio.TimeoutError(
            f"DB serialization failed after {max_attempts} retry transactions"
        )
    if result is Sentinel.TOKEN:
        raise DBOperationFailed("Transaction completed but no result was returned")
    return result


def create_async_engine(
    *args,
    _txn_concurrency_threshold: int = 0,
    _lock_conn_timeout: int = 0,
    **kwargs,
) -> ExtendedAsyncSAEngine:
    kwargs["future"] = True
    sync_engine = _create_engine(*args, **kwargs)
    return ExtendedAsyncSAEngine(
        sync_engine,
        _txn_concurrency_threshold=_txn_concurrency_threshold,
        _lock_conn_timeout=_lock_conn_timeout,
    )


@actxmgr
async def connect_database(
    db_config: DatabaseConfig,
    isolation_level: str = "SERIALIZABLE",
) -> AsyncIterator[ExtendedAsyncSAEngine]:
    from .base import pgsql_connect_opts

    db_url = (
        URL(f"postgresql+asyncpg://{db_config.addr.host}/{db_config.name}")
        .with_port(db_config.addr.port)
        .with_user(db_config.user)
    )
    if db_config.password is not None:
        db_url = db_url.with_password(db_config.password)

    version_check_db = create_async_engine(str(db_url))
    async with version_check_db.begin() as conn:
        result = await conn.execute(sa.text("show server_version"))
        version_str = result.scalar()
        major, minor, *_ = map(int, version_str.partition(" ")[0].split("."))
        if (major, minor) < (11, 0):
            pgsql_connect_opts["server_settings"].pop("jit")
    await version_check_db.dispose()

    db = create_async_engine(
        str(db_url),
        connect_args=pgsql_connect_opts,
        pool_size=db_config.pool_size,
        pool_recycle=db_config.pool_recycle,
        pool_pre_ping=db_config.pool_pre_ping,
        max_overflow=db_config.max_overflow,
        json_serializer=functools.partial(json.dumps, cls=ExtendedJSONEncoder),
        isolation_level=isolation_level,
        future=True,
        _txn_concurrency_threshold=max(
            int(db_config.pool_size + max(0, db_config.max_overflow) * 0.5),
            2,
        ),
        _lock_conn_timeout=int(db_config.lock_conn_timeout),
    )
    yield db
    await db.dispose()


@actxmgr
async def reenter_txn(
    pool: ExtendedAsyncSAEngine,
    conn: SAConnection,
    execution_opts: Mapping[str, Any] | None = None,
) -> AsyncIterator[SAConnection]:
    if conn is None:
        async with pool.connect() as conn:
            if execution_opts:
                await conn.execution_options(**execution_opts)
            async with conn.begin():
                yield conn
    else:
        async with conn.begin_nested():
            yield conn


@actxmgr
async def reenter_txn_session(
    pool: ExtendedAsyncSAEngine,
    sess: SASession,
    read_only: bool = False,
) -> AsyncIterator[SAConnection]:
    if sess is None:
        if read_only:
            async with pool.begin_readonly_session() as sess:
                yield sess
        else:
            async with pool.begin_session() as sess:
                yield sess
    else:
        async with sess.begin_nested():
            yield sess


# TODO: How to apply the missing execute_with_retry logic?
async def execute_with_retry(txn_func: Callable[[], Awaitable[TQueryResult]]) -> TQueryResult:
    max_attempts = 20
    result: TQueryResult | Sentinel = Sentinel.TOKEN
    try:
        async for attempt in AsyncRetrying(
            wait=wait_exponential(multiplier=0.02, min=0.02, max=5.0),
            stop=stop_after_attempt(max_attempts),
            retry=retry_if_exception_type(TryAgain),
        ):
            with attempt:
                try:
                    result = await txn_func()
                except DBAPIError as e:
                    if is_db_retry_error(e):
                        raise TryAgain
                    raise
    except RetryError:
        raise RuntimeError(f"DB serialization failed after {max_attempts} retries")
    if result is Sentinel.TOKEN:
        raise DBOperationFailed("Transaction completed but no result was returned")
    return result


async def retry_txn(max_attempts: int = 20) -> AsyncIterator[AttemptManager]:
    try:
        async for attempt in AsyncRetrying(
            wait=wait_exponential(multiplier=0.02, min=0.02, max=5.0),
            stop=stop_after_attempt(max_attempts),
            retry=retry_if_exception_type(TryAgain) | retry_if_exception_type(DBAPIError),
        ):
            # Since Python generators cannot catch the exceptions thrown in the code block executed
            # when yielded because stack frames are switched, we should pass AttemptManager to
            # provide a shared exception handling mechanism like the original execute_with_retry().
            yield attempt
            if attempt.retry_state.outcome is None:
                raise DBOperationFailed("Retry attempt completed but no outcome was recorded")
            exc = attempt.retry_state.outcome.exception()
            if isinstance(exc, DBAPIError) and not is_db_retry_error(exc):
                raise exc
    except RetryError:
        raise RuntimeError(f"DB serialization failed after {max_attempts} retries")


JSONCoalesceExpr: TypeAlias = sa.sql.elements.BinaryExpression


def sql_json_merge(
    col,
    key: Tuple[str, ...],
    obj: Mapping[str, Any],
    *,
    _depth: int = 0,
) -> JSONCoalesceExpr:
    """
    Generate an SQLAlchemy column update expression that merges the given object with
    the existing object at a specific (nested) key of the given JSONB column,
    with automatic creation of empty objects in parents and the target level.

    Note that the existing value must be also an object, not a primitive value.
    """
    expr = sa.func.coalesce(
        col if _depth == 0 else col[key[:_depth]],
        sa.text("'{}'::jsonb"),
    ).concat(
        (
            sa.func.jsonb_build_object(
                key[_depth],
                (
                    sa.func.coalesce(col[key], sa.text("'{}'::jsonb")).concat(
                        sa.func.cast(obj, psql.JSONB)
                    )
                    if _depth == len(key) - 1
                    else sql_json_merge(col, key, obj=obj, _depth=_depth + 1)
                ),
            )
            if key
            else sa.func.cast(obj, psql.JSONB)
        ),
    )
    return expr


def sql_json_increment(
    col,
    key: Tuple[str, ...],
    *,
    parent_updates: Optional[Mapping[str, Any]] = None,
    _depth: int = 0,
) -> JSONCoalesceExpr:
    """
    Generate an SQLAlchemy column update expression that increments the value at a specific
    (nested) key of the given JSONB column,
    with automatic creation of empty objects in parents and population of the
    optional parent_updates object to the target key's parent.

    Note that the existing value of the parent key must be also an object, not a primitive value.
    """
    expr = sa.func.coalesce(
        col if _depth == 0 else col[key[:_depth]],
        sa.text("'{}'::jsonb"),
    ).concat(
        sa.func.jsonb_build_object(
            key[_depth],
            (
                sa.func.coalesce(col[key].as_integer(), 0) + 1
                if _depth == len(key) - 1
                else sql_json_increment(col, key, parent_updates=parent_updates, _depth=_depth + 1)
            ),
        ),
    )
    if _depth == len(key) - 1 and parent_updates is not None:
        expr = expr.concat(sa.func.cast(parent_updates, psql.JSONB))
    return expr


def _populate_column(column: sa.Column):
    column_attrs = dict(column.__dict__)
    name = column_attrs.pop("name")
    return sa.Column(name, column.type, **{k: column_attrs[k] for k in column_constraints})


def regenerate_table(table: sa.Table, new_metadata: sa.MetaData) -> sa.Table:
    """
    This function can be used to regenerate table which belongs to SQLAlchemy ORM Class,
    which can be helpful when you're tring to build fresh new table for use on diffrent context
    than main manager logic (e.g. test code).
    Check out tests/test_image.py for more details.
    """
    return sa.Table(
        table.name,
        new_metadata,
        *[_populate_column(c) for c in table.columns],
    )


def agg_to_str(column: sa.Column) -> sa.sql.functions.Function:
    # https://docs.sqlalchemy.org/en/14/dialects/postgresql.html#sqlalchemy.dialects.postgresql.aggregate_order_by
    return sa.func.string_agg(column, psql.aggregate_order_by(sa.literal_column("','"), column))


def agg_to_array(column: sa.Column) -> sa.sql.functions.Function:
    return sa.func.array_agg(psql.aggregate_order_by(column, column.asc()))


def is_db_retry_error(e: Exception) -> bool:
    return isinstance(e, DBAPIError) and getattr(e.orig, "pgcode", None) == "40001"


async def vacuum_db(bootstrap_config: BootstrapConfig, vacuum_full: bool = False) -> None:
    async with connect_database(bootstrap_config.db, isolation_level="AUTOCOMMIT") as db:
        async with db.begin() as conn:
            vacuum_sql = "VACUUM FULL" if vacuum_full else "VACUUM"
            log.info(f"Perfoming {vacuum_sql} operation...")
            await conn.exec_driver_sql(vacuum_sql)



"""Base types and utilities for repository layer."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Callable, Generic, Optional, TypeVar

import sqlalchemy as sa
from sqlalchemy.engine import Row

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession as SASession

# QueryCondition now returns a ColumnElement (whereclause) instead of modifying stmt
type QueryCondition = Callable[[], sa.sql.expression.ColumnElement[bool]]

type QueryOrder = sa.sql.ClauseElement

# Factory function that creates a cursor condition from a decoded cursor value (str or UUID)
type CursorConditionFactory = Callable[[str], QueryCondition]


class QueryPagination(ABC):
    """
    Base class for pagination strategies.

    Subclasses must implement the apply() method to transform a SQLAlchemy
    select statement with appropriate pagination logic.
    """

    @property
    @abstractmethod
    def uses_window_function(self) -> bool:
        """Whether this pagination uses window function for total_count.

        Returns:
            True if window function should be added to main query (Offset),
            False if separate count query should be used (Cursor).
        """
        raise NotImplementedError

    @abstractmethod
    def apply(self, query: sa.sql.Select) -> sa.sql.Select:
        """Apply pagination to a SQLAlchemy select statement."""

        raise NotImplementedError

    @abstractmethod
    def compute_page_info(self, rows: list[Row], total_count: int) -> _PageInfoResult[Row]:
        """Compute pagination info and slice rows if needed.

        Args:
            rows: The rows returned from query (may include extra row for page detection)
            total_count: Total count of items matching the query

        Returns:
            _PageInfoResult containing sliced rows and pagination flags
        """

        raise NotImplementedError


@dataclass
class OffsetPagination(QueryPagination):
    """
    Offset-based pagination using limit and offset.

    This is the traditional SQL pagination approach where you specify:
    - limit: Maximum number of items to return
    - offset: Number of items to skip from the beginning
    """

    limit: int
    """Maximum number of items to return (must be positive)."""

    offset: int = 0
    """Number of items to skip from the beginning (must be non-negative)."""

    @property
    def uses_window_function(self) -> bool:
        return True

    def apply(self, query: sa.sql.Select) -> sa.sql.Select:
        """Apply offset-based pagination to query."""

        query = query.limit(self.limit)
        if self.offset > 0:
            query = query.offset(self.offset)
        return query

    def compute_page_info(self, rows: list[Row], total_count: int) -> _PageInfoResult[Row]:
        """Compute pagination info for offset-based pagination."""

        has_previous_page = self.offset > 0
        has_next_page = (self.offset + len(rows)) < total_count
        return _PageInfoResult(
            rows=rows,
            has_next_page=has_next_page,
            has_previous_page=has_previous_page,
        )


@dataclass
class CursorForwardPagination(QueryPagination):
    """
    Cursor-based forward pagination using first and after.

    This follows the GraphQL Relay Cursor Connections specification for forward pagination.
    Use this to paginate forward through a result set:
    - first: Number of items to return from the cursor position
    - cursor_condition: Optional QueryCondition for WHERE clause (e.g., id > cursor_id).
      If None, starts from the beginning.
    - cursor_order: QueryOrder for cursor-based ordering (e.g., created_at ASC)
    """

    first: int
    """Number of items to return (must be positive)."""

    cursor_order: QueryOrder
    """Ordering for cursor-based pagination (e.g., ORDER BY created_at ASC)."""

    cursor_condition: Optional[QueryCondition] = None
    """Optional QueryCondition for cursor position. If None, starts from the beginning."""

    @property
    def uses_window_function(self) -> bool:
        return False

    def apply(self, query: sa.sql.Select) -> sa.sql.Select:
        """
        Apply cursor-based forward pagination to query.

        Applies cursor condition (if present), cursor order, and LIMIT + 1 for has_next_page detection.
        """
        if self.cursor_condition is not None:
            query = query.where(self.cursor_condition())
        query = query.order_by(self.cursor_order)
        return query.limit(self.first + 1)

    def compute_page_info(self, rows: list[Row], total_count: int) -> _PageInfoResult[Row]:
        """Compute pagination info for cursor-based forward pagination."""

        # has_previous_page is True only if cursor was provided (meaning we're past the first page)
        has_previous_page = self.cursor_condition is not None
        has_next_page = len(rows) > self.first
        if has_next_page:
            rows = rows[: self.first]
        return _PageInfoResult(
            rows=rows,
            has_next_page=has_next_page,
            has_previous_page=has_previous_page,
        )


@dataclass
class CursorBackwardPagination(QueryPagination):
    """
    Cursor-based backward pagination using last and before.

    This follows the GraphQL Relay Cursor Connections specification for backward pagination.
    Use this to paginate backward through a result set:
    - last: Number of items to return before the cursor position
    - cursor_condition: Optional QueryCondition for WHERE clause (e.g., id < cursor_id).
      If None, starts from the end.
    - cursor_order: QueryOrder for cursor-based ordering (e.g., created_at DESC for reverse fetch)
    """

    last: int
    """Number of items to return (must be positive)."""

    cursor_order: QueryOrder
    """Ordering for cursor-based pagination (e.g., ORDER BY created_at DESC)."""

    cursor_condition: Optional[QueryCondition] = None
    """Optional QueryCondition for cursor position. If None, starts from the end."""

    @property
    def uses_window_function(self) -> bool:
        return False

    def apply(self, query: sa.sql.Select) -> sa.sql.Select:
        """
        Apply cursor-based backward pagination to query.

        Applies cursor condition (if present), cursor order, and LIMIT + 1 for has_previous_page detection.
        """
        if self.cursor_condition is not None:
            query = query.where(self.cursor_condition())
        query = query.order_by(self.cursor_order)
        return query.limit(self.last + 1)

    def compute_page_info(self, rows: list[Row], total_count: int) -> _PageInfoResult[Row]:
        """Compute pagination info for cursor-based backward pagination."""

        # has_next_page is True only if cursor was provided (meaning there are items after this page)
        has_next_page = self.cursor_condition is not None
        has_previous_page = len(rows) > self.last
        if has_previous_page:
            rows = rows[: self.last]
        return _PageInfoResult(
            rows=rows,
            has_next_page=has_next_page,
            has_previous_page=has_previous_page,
        )


@dataclass
class Querier:
    """Bundles query conditions, orders, and pagination for repository queries."""

    pagination: QueryPagination
    conditions: list[QueryCondition] = field(default_factory=list)
    orders: list[QueryOrder] = field(default_factory=list)


def combine_conditions_or(conditions: list[QueryCondition]) -> QueryCondition:
    """Combine multiple QueryConditions with OR logic.

    Args:
        conditions: List of QueryCondition callables to combine

    Returns:
        A single QueryCondition that applies all conditions with OR logic
    """

    def inner() -> sa.sql.expression.ColumnElement[bool]:
        clauses = [cond() for cond in conditions]
        return sa.or_(*clauses)

    return inner


def negate_conditions(conditions: list[QueryCondition]) -> QueryCondition:
    """Negate multiple QueryConditions with NOT logic.

    Args:
        conditions: List of QueryCondition callables to negate

    Returns:
        A single QueryCondition that negates the AND of all conditions
    """

    def inner() -> sa.sql.expression.ColumnElement[bool]:
        clauses = [cond() for cond in conditions]
        if len(clauses) == 1:
            return sa.not_(clauses[0])
        else:
            return sa.not_(sa.and_(*clauses))

    return inner


def _apply_querier(
    query: sa.sql.Select,
    querier: Querier,
) -> sa.sql.Select:
    """Apply query conditions, orders, and pagination to a SQLAlchemy select statement.

    Args:
        query: The base SELECT statement
        querier: Querier containing conditions, orders, and pagination to apply

    Returns:
        The modified SELECT statement with conditions, orders, and pagination applied

    Note:
        For cursor-based pagination, the pagination.apply() method applies
        cursor_order first. User-specified orders are applied after,
        serving as secondary sort criteria.
    """
    # Apply all conditions
    for condition in querier.conditions:
        query = query.where(condition())

    # Apply pagination (includes cursor condition and cursor_order for cursor pagination)
    query = querier.pagination.apply(query)

    # Apply user orders AFTER default order from pagination
    for order in querier.orders:
        query = query.order_by(order)

    return query


TRow = TypeVar("TRow", bound=Row)


@dataclass
class _PageInfoResult(Generic[TRow]):
    """Result of compute_page_info containing sliced rows and pagination flags."""

    rows: list[TRow]
    has_next_page: bool
    has_previous_page: bool


@dataclass
class QuerierResult(Generic[TRow]):
    """Result of executing a query with querier."""

    rows: list[TRow]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


async def execute_querier(
    db_sess: SASession,
    query: sa.sql.Select,
    querier: Querier,
) -> QuerierResult[Row]:
    """Execute query with querier and return rows with total_count and pagination info.

    For offset pagination, uses count().over() window function for efficient counting.
    For cursor pagination, executes a separate count query with filter conditions.

    Args:
        db_sess: Database session
        query: Base SELECT query (without count window function)
        querier: Querier for filtering, ordering, and pagination

    Returns:
        QuerierResult containing rows, total_count, and pagination info
    """
    initial_query = query

    # Add window function for offset pagination
    if querier.pagination.uses_window_function:
        query = query.add_columns(sa.func.count().over().label("total_count"))

    # Apply conditions and pagination to get data rows
    query = _apply_querier(query, querier)
    result = await db_sess.execute(query)
    rows = list(result.all())

    total_count: int
    if querier.pagination.uses_window_function and rows:
        # Offset pagination with results: use window function from rows
        total_count = rows[0].total_count
    else:
        # Cursor pagination or offset fallback (rows empty):
        # Execute pure count query with filter conditions only
        count_query = sa.select(sa.func.count()).select_from(initial_query.froms[0])
        for condition in querier.conditions:
            count_query = count_query.where(condition())

        count_result = await db_sess.execute(count_query)
        total_count = count_result.scalar() or 0

    # Calculate pagination info
    page_info = querier.pagination.compute_page_info(rows, total_count)

    return QuerierResult(
        rows=page_info.rows,
        total_count=total_count,
        has_next_page=page_info.has_next_page,
        has_previous_page=page_info.has_previous_page,
    )



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec

from .actions import (
    CreateChannelAction,
    CreateChannelActionResult,
    CreateRuleAction,
    CreateRuleActionResult,
    DeleteChannelAction,
    DeleteChannelActionResult,
    DeleteRuleAction,
    DeleteRuleActionResult,
    GetChannelAction,
    GetChannelActionResult,
    GetRuleAction,
    GetRuleActionResult,
    ProcessNotificationAction,
    ProcessNotificationActionResult,
    SearchChannelsAction,
    SearchChannelsActionResult,
    SearchRulesAction,
    SearchRulesActionResult,
    UpdateChannelAction,
    UpdateChannelActionResult,
    UpdateRuleAction,
    UpdateRuleActionResult,
    ValidateChannelAction,
    ValidateChannelActionResult,
    ValidateRuleAction,
    ValidateRuleActionResult,
)
from .service import NotificationService


class NotificationProcessors(AbstractProcessorPackage):
    """Processor package for notification operations."""

    create_channel: ActionProcessor[CreateChannelAction, CreateChannelActionResult]
    search_channels: ActionProcessor[SearchChannelsAction, SearchChannelsActionResult]
    get_channel: ActionProcessor[GetChannelAction, GetChannelActionResult]
    update_channel: ActionProcessor[UpdateChannelAction, UpdateChannelActionResult]
    delete_channel: ActionProcessor[DeleteChannelAction, DeleteChannelActionResult]
    validate_channel: ActionProcessor[ValidateChannelAction, ValidateChannelActionResult]

    create_rule: ActionProcessor[CreateRuleAction, CreateRuleActionResult]
    search_rules: ActionProcessor[SearchRulesAction, SearchRulesActionResult]
    get_rule: ActionProcessor[GetRuleAction, GetRuleActionResult]
    update_rule: ActionProcessor[UpdateRuleAction, UpdateRuleActionResult]
    delete_rule: ActionProcessor[DeleteRuleAction, DeleteRuleActionResult]
    validate_rule: ActionProcessor[ValidateRuleAction, ValidateRuleActionResult]

    process_notification: ActionProcessor[
        ProcessNotificationAction, ProcessNotificationActionResult
    ]

    def __init__(self, service: NotificationService, action_monitors: list[ActionMonitor]) -> None:
        self.create_channel = ActionProcessor(service.create_channel, action_monitors)
        self.search_channels = ActionProcessor(service.search_channels, action_monitors)
        self.get_channel = ActionProcessor(service.get_channel, action_monitors)
        self.update_channel = ActionProcessor(service.update_channel, action_monitors)
        self.delete_channel = ActionProcessor(service.delete_channel, action_monitors)
        self.validate_channel = ActionProcessor(service.validate_channel, action_monitors)

        self.create_rule = ActionProcessor(service.create_rule, action_monitors)
        self.search_rules = ActionProcessor(service.search_rules, action_monitors)
        self.get_rule = ActionProcessor(service.get_rule, action_monitors)
        self.update_rule = ActionProcessor(service.update_rule, action_monitors)
        self.delete_rule = ActionProcessor(service.delete_rule, action_monitors)
        self.validate_rule = ActionProcessor(service.validate_rule, action_monitors)

        self.process_notification = ActionProcessor(service.process_notification, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateChannelAction.spec(),
            SearchChannelsAction.spec(),
            GetChannelAction.spec(),
            UpdateChannelAction.spec(),
            DeleteChannelAction.spec(),
            ValidateChannelAction.spec(),
            CreateRuleAction.spec(),
            SearchRulesAction.spec(),
            GetRuleAction.spec(),
            UpdateRuleAction.spec(),
            DeleteRuleAction.spec(),
            ValidateRuleAction.spec(),
            ProcessNotificationAction.spec(),
        ]



"""add kernel last seen

Revision ID: 42998e8c8bb4
Revises: c7fedaca4883
Create Date: 2025-05-13 15:42:29.976150

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "42998e8c8bb4"
down_revision = "c7fedaca4883"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "kernels",
        sa.Column(
            "last_seen", sa.DateTime(timezone=True), server_default=sa.text("NULL"), nullable=True
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "last_seen")
    # ### end Alembic commands ###



from __future__ import annotations

import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Self, Sequence

import graphene
import sqlalchemy as sa
from dateutil.parser import parse as dtparse
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.engine.row import Row

from ai.backend.common.clients.valkey_client.valkey_rate_limit.client import ValkeyRateLimitClient
from ai.backend.common.defs import REDIS_RATE_LIMIT_DB, RedisRole
from ai.backend.common.types import AccessKey
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.keypair.types import KeyPairCreator, KeyPairData
from ai.backend.manager.models.gql_models.session import ComputeSession
from ai.backend.manager.models.keypair import (
    keypairs,
    prepare_new_keypair,
)

if TYPE_CHECKING:
    from ..gql import GraphQueryContext
    from .vfolder import VirtualFolder

__all__ = (
    "UserInfo",
    "KeyPair",
    "KeyPairInput",
    "KeyPairList",
    "ModifyKeyPairInput",
    "CreateKeyPair",
    "ModifyKeyPair",
    "DeleteKeyPair",
)

from ..base import (
    Item,
    PaginatedList,
    batch_multiresult,
    batch_result,
    set_if_set,
    simple_db_mutate,
    simple_db_mutate_returning_item,
)
from ..minilang.ordering import OrderSpecItem, QueryOrderParser
from ..minilang.queryfilter import FieldSpecItem, QueryFilterParser
from ..user import UserRole
from ..utils import agg_to_array


class UserInfo(graphene.ObjectType):
    email = graphene.String()
    full_name = graphene.String()

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> Optional[UserInfo]:
        if row is None:
            return None
        return cls(email=row["email"], full_name=row["full_name"])

    @classmethod
    async def batch_load_by_uuid(
        cls,
        ctx: GraphQueryContext,
        user_uuids: Sequence[uuid.UUID],
    ) -> Sequence[Optional[UserInfo]]:
        async with ctx.db.begin_readonly() as conn:
            from .user import users

            query = (
                sa.select([users.c.uuid, users.c.email, users.c.full_name])
                .select_from(users)
                .where(users.c.uuid.in_(user_uuids))
            )
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                user_uuids,
                lambda row: row["uuid"],
            )


class KeyPair(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    user_id = graphene.String()
    full_name = graphene.String()
    access_key = graphene.String()
    secret_key = graphene.String()
    is_active = graphene.Boolean()
    is_admin = graphene.Boolean()
    resource_policy = graphene.String()
    created_at = GQLDateTime()
    last_used = GQLDateTime()
    rate_limit = graphene.Int()
    num_queries = graphene.Int()
    rolling_count = graphene.Int(description="Added in 24.09.0.")
    user = graphene.UUID()
    projects = graphene.List(lambda: graphene.String)

    ssh_public_key = graphene.String()

    vfolders = graphene.List("ai.backend.manager.models.VirtualFolder")
    compute_sessions = graphene.List(
        ComputeSession,
        status=graphene.String(),
    )
    concurrency_used = graphene.Int()

    user_info = graphene.Field(lambda: UserInfo)

    # Deprecated
    concurrency_limit = graphene.Int(
        deprecation_reason=(
            "Moved to KeyPairResourcePolicy object as the max_concurrent_sessions field."
        )
    )

    async def resolve_user_info(
        self,
        info: graphene.ResolveInfo,
    ) -> UserInfo:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, "UserInfo.by_uuid")
        return await loader.load(self.user)

    @classmethod
    def from_data(cls, data: KeyPairData) -> Self:
        return cls(
            access_key=data.access_key,
            secret_key=data.secret_key,
            is_active=data.is_active,
            is_admin=data.is_admin,
            resource_policy=data.resource_policy_name,
            created_at=data.created_at,
            rate_limit=data.rate_limit,
            user=data.user_id,
            ssh_public_key=data.ssh_public_key,
        )

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> KeyPair:
        return cls(
            id=row["access_key"],
            user_id=row["user_id"],
            full_name=row["full_name"] if "full_name" in row.keys() else None,
            access_key=row["access_key"],
            secret_key=row["secret_key"],
            is_active=row["is_active"],
            is_admin=row["is_admin"],
            resource_policy=row["resource_policy"],
            created_at=row["created_at"],
            rate_limit=row["rate_limit"],
            user=row["user"],
            ssh_public_key=row["ssh_public_key"],
            concurrency_limit=0,  # deprecated
            projects=row["groups_name"] if "groups_name" in row.keys() else [],
        )

    async def resolve_num_queries(self, info: graphene.ResolveInfo) -> int:
        ctx: GraphQueryContext = info.context
        return await ctx.valkey_stat.get_keypair_query_count(self.access_key)

    async def resolve_rolling_count(self, info: graphene.ResolveInfo) -> int:
        ctx: GraphQueryContext = info.context
        valkey_profile_target = ctx.config_provider.config.redis.to_valkey_profile_target()
        valkey_target = valkey_profile_target.profile_target(RedisRole.RATE_LIMIT)
        valkey_client = await ValkeyRateLimitClient.create(
            valkey_target=valkey_target,
            db_id=REDIS_RATE_LIMIT_DB,
            human_readable_name="ratelimit",
        )
        try:
            return await valkey_client.get_rolling_count(self.access_key)
        finally:
            await valkey_client.close()

    async def resolve_vfolders(self, info: graphene.ResolveInfo) -> Sequence[VirtualFolder]:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, "VirtualFolder")
        return await loader.load(self.access_key)

    async def resolve_compute_sessions(
        self, info: graphene.ResolveInfo, raw_status: Optional[str] = None
    ):
        ctx: GraphQueryContext = info.context

        if raw_status is not None:
            status = KernelStatus[raw_status]
        loader = ctx.dataloader_manager.get_loader(ctx, "ComputeSession", status=status)
        return await loader.load(self.access_key)

    async def resolve_concurrency_used(self, info: graphene.ResolveInfo) -> int:
        ctx: GraphQueryContext = info.context

        # Get repository from context
        repository = ctx.scheduler_repository

        # Get concurrency through repository (cache-through pattern)
        # Convert graphene.String to str, then to AccessKey type
        access_key = AccessKey(str(self.access_key))
        return await repository.get_keypair_concurrency(access_key, is_sftp=False)

    async def resolve_last_used(self, info: graphene.ResolveInfo) -> datetime | None:
        ctx: GraphQueryContext = info.context
        row_ts = await ctx.valkey_stat.get_keypair_last_used_time(self.access_key)
        if row_ts is None:
            return None
        return datetime.fromtimestamp(row_ts)

    @classmethod
    async def load_all(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        is_active: Optional[bool] = None,
        limit: Optional[int] = None,
    ) -> Sequence[KeyPair]:
        from .user import users

        j = sa.join(
            keypairs,
            users,
            keypairs.c.user == users.c.uuid,
        )
        query = sa.select([keypairs]).select_from(j)
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        if limit is not None:
            query = query.limit(limit)
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, row)) is not None
            ]

    _queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
        "access_key": ("keypairs_access_key", None),
        "user_id": ("users_uuid", None),
        "email": ("users_email", None),
        "full_name": ("users_full_name", None),
        "is_active": ("keypairs_is_active", None),
        "is_admin": ("keypairs_is_admin", None),
        "resource_policy": ("keypairs_resource_policy", None),
        "created_at": ("keypairs_created_at", dtparse),
        "last_used": ("keypairs_last_used", dtparse),
        "rate_limit": ("keypairs_rate_limit", None),
        "num_queries": ("keypairs_num_queries", None),
        "ssh_public_key": ("keypairs_ssh_public_key", None),
        "projects": ("groups_name", None),
    }

    _queryorder_colmap: Mapping[str, OrderSpecItem] = {
        "access_key": ("keypairs_access_key", None),
        "email": ("users_email", None),
        "full_name": ("users_full_name", None),
        "is_active": ("keypairs_is_active", None),
        "is_admin": ("keypairs_is_admin", None),
        "resource_policy": ("keypairs_resource_policy", None),
        "created_at": ("keypairs_created_at", None),
        "last_used": ("keypairs_last_used", None),
        "rate_limit": ("keypairs_rate_limit", None),
        "num_queries": ("keypairs_num_queries", None),
        "projects": ("groups_name", agg_to_array),
    }

    @classmethod
    async def load_count(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        email: Optional[str] = None,
        is_active: Optional[bool] = None,
        filter: Optional[str] = None,
    ) -> int:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .outerjoin(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .outerjoin(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = sa.select([sa.func.count()]).group_by(keypairs.c.access_key).select_from(j)
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if email is not None:
            query = query.where(keypairs.c.user_id == email)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with graph_ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return len(result.all())

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: Optional[str] = None,
        email: Optional[str] = None,
        is_active: Optional[bool] = None,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence[KeyPair]:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .outerjoin(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .outerjoin(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = (
            sa.select([
                keypairs,
                users.c.email,
                users.c.full_name,
                agg_to_array(groups.c.name).label("groups_name"),
            ])
            .select_from(j)
            .group_by(keypairs, users.c.email, users.c.full_name)
            .limit(limit)
            .offset(offset)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if email is not None:
            query = query.where(keypairs.c.user_id == email)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(keypairs.c.created_at.desc())
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, row)) is not None
            ]

    @classmethod
    async def batch_load_by_email(
        cls,
        graph_ctx: GraphQueryContext,
        user_ids: Sequence[uuid.UUID],
        *,
        domain_name: Optional[str] = None,
        is_active: Optional[bool] = None,
    ) -> Sequence[Sequence[Optional[KeyPair]]]:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .join(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .join(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = (
            sa.select([
                keypairs,
                users.c.email,
                users.c.full_name,
                agg_to_array(groups.c.name).label("groups_name"),
            ])
            .select_from(j)
            .where(keypairs.c.user_id.in_(user_ids))
            .group_by(keypairs, users.c.email, users.c.full_name)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_multiresult(
                graph_ctx,
                conn,
                query,
                cls,
                user_ids,
                lambda row: row["user_id"],
            )

    @classmethod
    async def batch_load_by_ak(
        cls,
        graph_ctx: GraphQueryContext,
        access_keys: Sequence[AccessKey],
        *,
        domain_name: Optional[str] = None,
    ) -> Sequence[Optional[KeyPair]]:
        from .group import association_groups_users, groups
        from .user import users

        j = (
            sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
            .join(association_groups_users, users.c.uuid == association_groups_users.c.user_id)
            .join(groups, association_groups_users.c.group_id == groups.c.id)
        )
        query = (
            sa.select([
                keypairs,
                users.c.email,
                users.c.full_name,
                agg_to_array(groups.c.name).label("groups_name"),
            ])
            .select_from(j)
            .where(keypairs.c.access_key.in_(access_keys))
            .group_by(keypairs, users.c.email, users.c.full_name)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_result(
                graph_ctx,
                conn,
                query,
                cls,
                access_keys,
                lambda row: row["access_key"],
            )


class KeyPairList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(KeyPair, required=True)


class KeyPairInput(graphene.InputObjectType):
    is_active = graphene.Boolean(required=False, default_value=True)
    is_admin = graphene.Boolean(required=False, default_value=False)
    resource_policy = graphene.String(required=True)
    concurrency_limit = graphene.Int(required=False)  # deprecated and ignored
    rate_limit = graphene.Int(required=True)

    # When creating, you MUST set all fields.
    # When modifying, set the field to "None" to skip setting the value.

    def to_creator(self) -> KeyPairCreator:
        return KeyPairCreator(
            is_active=self.is_active,
            is_admin=self.is_admin,
            resource_policy=self.resource_policy,
            rate_limit=self.rate_limit,
        )


class ModifyKeyPairInput(graphene.InputObjectType):
    is_active = graphene.Boolean(required=False)
    is_admin = graphene.Boolean(required=False)
    resource_policy = graphene.String(required=False)
    concurrency_limit = graphene.Int(required=False)  # deprecated and ignored
    rate_limit = graphene.Int(required=False)


class CreateKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        user_id = graphene.String(required=True)
        props = KeyPairInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    keypair = graphene.Field(lambda: KeyPair, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        user_id: str,
        props: KeyPairInput,
    ) -> CreateKeyPair:
        from .user import users  # noqa

        graph_ctx: GraphQueryContext = info.context
        data = prepare_new_keypair(user_id, props.to_creator())
        insert_query = sa.insert(keypairs).values(
            **data,
            user=sa.select([users.c.uuid]).where(users.c.email == user_id).as_scalar(),
        )
        return await simple_db_mutate_returning_item(cls, graph_ctx, insert_query, item_cls=KeyPair)


class ModifyKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        access_key = graphene.String(required=True)
        props = ModifyKeyPairInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        access_key: AccessKey,
        props: ModifyKeyPairInput,
    ) -> ModifyKeyPair:
        ctx: GraphQueryContext = info.context
        data: Dict[str, Any] = {}
        set_if_set(props, data, "is_active")
        set_if_set(props, data, "is_admin")
        set_if_set(props, data, "resource_policy")
        set_if_set(props, data, "rate_limit")
        # props.concurrency_limit is always ignored
        update_query = sa.update(keypairs).values(data).where(keypairs.c.access_key == access_key)
        return await simple_db_mutate(cls, ctx, update_query)


class DeleteKeyPair(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        access_key = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        access_key: AccessKey,
    ) -> DeleteKeyPair:
        from .user import UserRow

        ctx: GraphQueryContext = info.context
        async with ctx.db.begin_readonly_session() as db_session:
            user_query = (
                sa.select([sa.func.count()])
                .select_from(UserRow)
                .where(UserRow.main_access_key == access_key)
            )
            if (await db_session.scalar(user_query)) > 0:
                return DeleteKeyPair(False, "the keypair is used as main access key by any user")
        delete_query = sa.delete(keypairs).where(keypairs.c.access_key == access_key)
        result = await simple_db_mutate(cls, ctx, delete_query)
        if result.ok:
            await ctx.valkey_stat.delete_keypair_concurrency(
                access_key=str(access_key),
                is_private=False,
            )
        return result



"""
Common and generic exceptions that don't belong to a specific domain.
"""

from __future__ import annotations

from typing import Any, Optional

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.common.plugin.hook import HookResult


class URLNotFound(BackendAIError, web.HTTPNotFound):  # TODO: Misused now.
    error_type = "https://api.backend.ai/probs/url-not-found"
    error_title = "Unknown URL path."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ObjectNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/object-not-found"
    object_name = "object"

    def __init__(
        self,
        extra_msg: Optional[str] = None,
        extra_data: Optional[Any] = None,
        *,
        object_name: Optional[str] = None,
        **kwargs,
    ) -> None:
        if object_name:
            self.object_name = object_name
        self.error_title = f"No such {self.object_name}."
        super().__init__(extra_msg, extra_data, **kwargs)

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class GenericBadRequest(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Bad request."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class GenericForbidden(BackendAIError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/generic-forbidden"
    error_title = "Forbidden operation."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class RejectedByHook(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/rejected-by-hook"
    error_title = "Operation rejected by a hook plugin."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.PLUGIN,
            operation=ErrorOperation.HOOK,
            error_detail=ErrorDetail.FORBIDDEN,
        )

    @classmethod
    def from_hook_result(cls, result: HookResult) -> RejectedByHook:
        return cls(
            extra_msg=result.reason,
            extra_data={
                "plugins": result.src_plugin,
            },
        )


class MethodNotAllowed(BackendAIError, web.HTTPMethodNotAllowed):
    error_type = "https://api.backend.ai/probs/method-not-allowed"
    error_title = "HTTP Method Not Allowed."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class InternalServerError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Internal server error."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ServerMisconfiguredError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/server-misconfigured"
    error_title = "Service misconfigured."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ServiceUnavailable(BackendAIError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/service-unavailable"
    error_title = "Serivce unavailable."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class ServerFrozen(BackendAIError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/server-frozen"
    error_title = "The server is frozen due to maintenance. Please try again later."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class Forbidden(BackendAIError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/forbidden"
    error_title = "Forbidden operation."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.FORBIDDEN,
        )



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from zmq.auth.certs import load_certificate

from ai.backend.common.auth import PublicKey, SecretKey
from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.manager.agent_cache import AgentRPCCache
from ai.backend.manager.config.unified import ManagerUnifiedConfig
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

from ..errors import InvalidManagerKeypairError


@dataclass
class AgentCacheInput:
    """Input required for agent cache setup.

    Contains database and RPC authentication configuration.
    """

    db: ExtendedAsyncSAEngine
    config: ManagerUnifiedConfig


class AgentCacheDependency(NonMonitorableDependencyProvider[AgentCacheInput, AgentRPCCache]):
    """Provides AgentRPCCache lifecycle management."""

    @property
    def stage_name(self) -> str:
        return "agent-cache"

    @asynccontextmanager
    async def provide(self, setup_input: AgentCacheInput) -> AsyncIterator[AgentRPCCache]:
        """Initialize and provide agent RPC cache.

        Args:
            setup_input: Input containing database and config

        Yields:
            Initialized AgentRPCCache
        """
        # Load manager keypair for RPC authentication
        manager_pkey, manager_skey = load_certificate(
            setup_input.config.manager.rpc_auth_manager_keypair
        )
        if manager_skey is None:
            raise InvalidManagerKeypairError("Manager secret key is missing from the keypair file")
        manager_public_key = PublicKey(manager_pkey)
        manager_secret_key = SecretKey(manager_skey)

        # Create agent cache
        agent_cache = AgentRPCCache(setup_input.db, manager_public_key, manager_secret_key)

        yield agent_cache
